%%%-------------------------------------------------------------------
%%% File    : berkeley_db_replication_group_helpers.hrl
%%% Author  : Chris Newcombe <>
%%% Description : 
%%%
%%% Created : 26 Oct 2006 by Chris Newcombe <>
%%%-------------------------------------------------------------------

%% This record holds an approximation of the 'live state'
-record(berkeley_db_replication_group_caller_state, 

        {replication_group_name,

         all_site_locators,

         %% TODO: If a group is large, we might want to limit the 
         %% number of sites we try before giving up.
         %% This is a constant config parameter.
         %%   max_num_sites_to_try_per_request, 

         up_site_locators,

         down_sites,                 % [{SiteLocator, LastPolledTime}, ...]

         %% If we fail to connect to a site that is in UpSiteLocators
         %% then we move it to DownSites, and record the time that it
         %% moved there.  We don't normally try the site again for
         %% this period of time (see below for exception).
         %% Therefore this is an estimate of the recovery time for a 
         %% site (MTTR might be too small; might want 70th percentile or so).
         %% TODO: add expoential back-off to this, up to some limit.
         %% This is a constant config parameter.
         ignore_down_sites_for_milliseconds,

         %% Exception: If ALL sites are down, then we will start to
         %% probe random sites (call ?BDB_RG:get_status) to see if any
         %% recover earlier than this timeout, as this timeout is probably 
         %% conservative.  If we do find such a site then we update 
         %% up_site_locators & down_sites with that site's view of the 
         %% entire group's status, as that is presumed to be more 
         %% up to date than the client's view.
         %% TODO
         %% Currently while all sites are down, *every* request 
         %% will attempt to poll 
         %% If a new request finds that up_site_locators is empty 
         %% (i.e. all sites are in down_sites)
         %% then to doing too much work (all parallel requests probing 
         %% sites for changes) then we could and probably should throttle 
         %% the probes
         %%    last_poll_when_all_sites_down_time
         %%    poll_when_all_sites_down_period_milliseconds

         %% Obviously the client caller-state holds it's current idea of 
         %t which site is master
         %%
         %%  - #master_location_hint record of a site in up_site_locators
         %%    (this is currently a SiteLocator and a Distrbuted Erlang pid of the replication_group server, for faster access)
         %%
         %%  - 'unknown'
         %% 
         %% If ?BDB_RG_H:do_txn_on_master is called when it is 'unknown' then we
         %% call get_master_location_hint (with the client-specified timeout), so that we minimize 
         %% the number of failed requests exposed to clients during failovers.
         
         master_location_hint,

         %% This enire record is typically stored in an ETS table per node,
         %% and multiple clients use their own copies of it concurrently.
         %% If a client discovers a change -- e.g. a site is down, 
         %% or the master has moved, it will use persis_write_fun
         %% to store the updated version.
         %% However, other clients may have noticed other changes 
         %% and stored a new version that is more recent than 
         %% the version that this client began with 
         %% (e.g. another client discovered that the master 
         %% moved and wrote that change, but this client only 
         %% discovered that some site is down).
         %% To avoid overwriting changes, we use last_persisted_time
         %% to reject writes based on versions earlier than 
         %% the current persistent version.
         %% Eventually clients could merge versions,
         %% but currently they just drop the write.
         %% The idea is that eventually some client will win.

         last_persisted_time,
         persist_read_fun,
         persist_write_fun
        }).
