
# patches-to-berkeley-db/db-4.5.20/apply-patches.sh
#
# Run this:
#
#    cd /home/$USER/erlang/downloads/edtk
#    tar -zxvf db-4.5.20.tar.gz
#    cd db-4.5.20
#    sh ../edtk-1.5/examples/berkeley_db/patches-to-berkeley-db/db-4.5.20/apply-patches.sh \
#      ../edtk-1.5/examples/berkeley_db/patches-to-berkeley-db/db-4.5.20
#    cd build_unix
#    ../dist/configure --enable-debug --prefix=`realpath ../../BerkeleyDB.4.5`
#    make
#    make install

# Best guess as to where to find the patches, but we can override it
# by passing in a command line argument.
patch_dir=${1:-../edtk-1.5/examples/berkeley_db/patches-to-berkeley-db/db-4.5.20}


# This super-patch is a Sleepycat-generated equivalent to the patches listed at the end of this file.
patch -p1 <$patch_dir/20070108.01

# [#15135] Fix a bug that could cause a checkpoint to hang if a database was closed while the checkpoint was forcing that file to disk and all the pages for that database were replaced in the cache.
patch -p1 <$patch_dir/20070124.01

# [#15223] Fix crash in Java interface optimizations for replication
# (Patches below assume that we are not using this one)
###patch -p1 <$patch_dir/20070125.01

# [#14752] Fix for hang in previous patch (now in super-patch 20070108.01) for election durability issue (suspend writes during election)
patch -p1 <$patch_dir/20070126.01

# [#15342] Fix for underflow of "Current clean page count" in db_stat -m output, which might also have caused memp_trickle to consume lots of cpu cycles
#          Also, mp_set_max_write now takes effect immediately even if a checkpoing/memp_trickle/memp_sync is in progress. This allows faster shutdown 
#          of environments as throttled dirty-page operations might take minutes to complete on large caches.
patch -p1 <$patch_dir/20070201.01

# [#15338] Fixes for replicated writes being blocked due to delayed IS_PERM acks when a checkpoint is in progress and mp_max_write (throttled dirty-page IO) is used
#          (master now sends advance warning to replicas that a checkpoint is starting so that replicas can do the necessary IO in parallel with the master)
patch -p1 <$patch_dir/20070202.01
patch -p1 <$patch_dir/20070209.01
patch -p1 <$patch_dir/20070215.01

# [#15342] or possibly [#15338] Fix for incorrect dirty-page count in db_stat -l
patch -p1 <$patch_dir/20070219.01

# [#15338] (more) Fix for replicated writes being blocked due to delayed IS_PERM acks when a checkpoint is in progress and mp_max_write (throttled dirty-page IO) is used
#          (txn_checkpoint nolonger flushes dirty cache pages that belong to __db_rep.db)
echo Applying 20070228.01
patch -p1 <$patch_dir/20070228.01

echo Applying 20070314.01
# [#15338] DB_EVENT_REP_MASTER is nolonger delayed by master-sleep in txn_checkpoint
patch -p1 <$patch_dir/20070314.01

echo Applying 20070319.01
# [#15338] temporary hack to allow configuration of master-sleep duration in txn_checkpoint
patch -p1 <$patch_dir/20070319.01

echo Applying 20070321.01
# [#15338] fix replication stall due to logfile changeover (newfile)
patch -p1 <$patch_dir/20070321.01

echo Applying 20070327.01
# [#15338] when a checkpoint record arrives on a replica, temporarily disable a client's mp_max_write so that the startsync can quickly complete.
patch -p1 <$patch_dir/20070327.01

echo Applying 20070404.01
# [#15428] endless elections (and therfore endless PERM_FAILED) in some circumstances
patch -p1 <$patch_dir/20070404.01

echo Applying 20070405.01
# [#15434] maxwrite is checked while holding a hash bucket locked 
patch -p1 <$patch_dir/20070405.01

echo Applying 20070406.01
# [#14754] txn_recover() does not work correctly after DB_ENV_REP_MASTER is received on a failover master
patch -p1 <$patch_dir/20070406.01

echo Applying 20070425.01
# [#14754] txn_recover() does not work correctly after DB_ENV_REP_MASTER is received on a failover master
patch -p1 <$patch_dir/20070425.01


# Our own unofficial patches

echo Applying ---- interrupt_os_sleep.patch ----
# [#15338] hack to allow interruption of long os_sleep calls (allow fast shutdown if master-sleep is set to tens/hundreds of seconds)
patch -p1 <$patch_dir/interrupt_os_sleep.patch

# [#15342] Fix for corrupt MVCC stats in db_stat -m output, if MVCC is not enabled (cosmetic)
###patch -p0 <$patch_dir/4.5-zero-mvcc-stats.diff


# Earlier patches incorporated into first patch above (20070108.01) :
#
## The order of these is important.
#
#  # [DB_VERB_REPLICATION requiring --enable-diagnostic]
#  patch -p0 <$patch_dir/patch.14991
#   
#  # [temporary but critical hack: manual method to prevent Db:close from logging, for use if there are prepared txns]
#  patch -p0 <$patch_dir/patch.14785
#   
#  # [continuing bulk transfer after throttling]
#  patch -p1 <$patch_dir/patch.15136
#   
#  # [txn create on rep client disallowed]
#  patch -p1 <$patch_dir/patch.15146
#   
#  # [Internal init and elections]
#  patch -p0 <$patch_dir/patch.15160
#   
#  # [Election strangeness]
#  patch -p1 <$patch_dir/patch.15186
#   
#  # [replica-to-replica sync]
#  patch -p1 <$patch_dir/patch.15234
#  patch -p1 <$patch_dir/patch.15234.2
#   
#  # [implements DB_EVENT_REP_PERM_FAILED]
#  patch -p1 <$patch_dir/patch.14775
#   
#  # [lock_detect does not count time-expired locks/txns in *abortp]
#  patch -p0 <$patch_dir/patch.15281
#   
#  # [Election strangeness #2]
#  patch -p1 <$patch_dir/patch.15186.2
#   
#  # [minor repmgr memory leaks]
#  patch -p1 <$patch_dir/patch.15256
#   
#  # [repmgr connection drops occasinally prevent elections]   [FAILS to patch rep_windows.c] 
#  patch -p1 <$patch_dir/patch.15311
#   
#  # # [Durability of transactions when an election is running]
#  patch -p1 <$patch_dir/patch.14752.3
#  patch -p1 <$patch_dir/patch.14752.4
