#!/bin/bash

# make_release.sh for berkeley_db driver
#
# Must be run in its own directory (examples/berkeley_db_releases).
# Creates an Erlang/OTP application in a named subdirectory of the current directory.
#
# Example usage:
#
#    pushd ~/erlang/downloads/edtk/edtk-1.5/examples/berkeley_db/releases; ./make_release.sh berkeley_db-1.0; popd
#
# IMPORTANT (you may need to edit some lines):
#
#   This assumes that BDB (*plus patches* -- see README-cnewcom) is installed in:
#
#       ../../../../BerkeleyDB.4.5    (i.e. the same location that examples/berkeley_db/Makefile assumes)
#
#   and that http://goog-coredumper.sourceforge.net/ is installed in:
#
#       /usr/local/lib
#
# See the lines marked "MAY NEED TO CHANGE THIS" below.
#

set -e

if [[ -d $1 ]]; then
    echo "$1 already exists - aborting."
    exit 1
fi

echo "Doing a clean build of EDTK and the berkeley_db driver"

# Note we actually re-build all of the .erl files after copying them 
# (to set +debug_info).  But we do a clean build to check for 
# build errors, and to make sure that berkeley_db_drv.so matches 
# the .erl files that we copy.

pushd ../../../edtk
make clean
make
popd

pushd ..
make clean
make  
popd

echo "Copying release files to $1"

mkdir $1
mkdir $1/src
mkdir $1/include
mkdir $1/ebin
mkdir $1/priv

cp ../berkeley_db_drv.hrl                                 $1/include
cp ../berkeley_db_replication_group_helpers.hrl           $1/include
cp ../berkeley_db_replication_group_internal.hrl          $1/include

cp ../berkeley_db_drv.erl                                 $1/src
cp ../berkeley_db_utilities.erl                           $1/src
cp ../berkeley_db_port_coordinator.erl                    $1/src
cp ../berkeley_db_helpers.erl                             $1/src
cp ../berkeley_db_sequence_server.erl                     $1/src
cp ../berkeley_db_global_transaction_manager.erl          $1/src
cp ../berkeley_db_replication_group.erl                   $1/src
cp ../berkeley_db_replication_group_gen_event_handler.erl $1/src
cp ../berkeley_db_replication_group_helpers.erl           $1/src

cp ../berkeley_db_drv.so                                  $1/priv
cp ../../../edtk/pipe-main                                $1/priv

cp ../berkeley_db.app                                     $1/ebin

# These lines depend on installation paths of other packages
# MAY NEED TO CHANGE THIS
cp  /usr/local/lib/libcoredumper.so                       $1/priv
cp  /usr/local/lib/libpcre.so                            $1/priv
cp ../../../../BerkeleyDB.4.5/lib/libdb-4.5.so            $1/priv
cp ../../../../BerkeleyDB.4.5/bin/db_recover              $1/priv
cp ../../../../BerkeleyDB.4.5/bin/db_stat                 $1/priv
cp ../../../../BerkeleyDB.4.5/bin/db_archive              $1/priv
cp ../../../../BerkeleyDB.4.5/bin/db_hotbackup            $1/priv
cp ../../../../BerkeleyDB.4.5/bin/db_printlog             $1/priv
cp ../../../../BerkeleyDB.4.5/bin/db_verify               $1/priv
cp ../../../../BerkeleyDB.4.5/bin/db_dump                 $1/priv
cp ../../../../BerkeleyDB.4.5/bin/db_load                 $1/priv

echo "Compiling $1/src/*.erl to $1/ebin/*.beam"
pushd $1/src
erlc -W +debug_info -I ../include -o ../ebin *.erl
popd

echo "Done."
