%%%----------------------------------------------------------------------
%%% File    : ethbr_iface_sup.erl
%%% Author  : Scott Lystig Fritchie <slfritchie@snookles.com>
%%% Purpose : Ethernet bridge interface supervisor
%%% Created : 01 Aug 2002 by Scott Lystig Fritchie <slfritchie@snookles.com>
%%%----------------------------------------------------------------------

-module(ethbr_iface_sup).
-compile([verbose, report_errors, report_warnings, trace, debug_info]).

-behaviour(supervisor).

%% External exports
-export([start_link/0, start_link/1]).

%% supervisor callbacks
-export([init/1]).

-export([xxx/1]).
xxx(Args) ->
    io:format("XXX ~w:xxx() Args = ~p\n", [?MODULE, Args]),
    {ok, self()}.

%%%----------------------------------------------------------------------
%%% API
%%%----------------------------------------------------------------------
start_link() ->
    io:format("XXX ~w:start_link()\n", [?MODULE]),
    {error, "Must include StartArgs list"}.
start_link(StartArgs) ->
    io:format("XXX ~w:start_link ArgList = ~w\n", [?MODULE, StartArgs]),
    case lists:keysearch(iface_sup_name, 1, StartArgs) of
	{value, {_, ISName}} ->
	    supervisor:start_link({local, ISName}, ?MODULE, StartArgs);
	_ ->
	    {error, {iface_sup_name, not_found}}
    end.

%%%----------------------------------------------------------------------
%%% Callback functions from supervisor
%%%----------------------------------------------------------------------

%%----------------------------------------------------------------------
%% Func: init/1
%% Returns: {ok,  {SupFlags,  [ChildSpec]}} |
%%          ignore                          |
%%          {error, Reason}   
%%----------------------------------------------------------------------
init(ArgList) ->
    io:format("XXX ~w:init ArgList = ~w\n", [?MODULE, ArgList]),
    %% Contact center, let it know we're here.  I'd rather not do this,
    %% but I don't know another way around it.  I want the ethbridge_sup
    %% to use rest_for_one restart strategy, and I want this iface sup
    %% to the "right" of the center proc, so if center proc dies, we die,
    %% but if we die, center proc doesn't (normally) die.
    %% However, center must tell us how to start new children for local
    %% interfaces, but center doesn't know we're up yet.  So, we'll send
    %% center a cast to let it know that we're alive.
    %%
    %% XXX There may yet be a race condition where center receives the
    %% cast and starts connecting interfaces but we are not yet registered.
    %% If we send our pid in the cast, perhaps someone else could be
    %% impersonating us?  But if there is a race, then if center always
    %% contacts us by registered name, there's a window where it couldn't
    %% contact us?
    %%
    %% QQQ You know, this begs a question, and I don't know the answer.
    %% Why does the ethbr_iface_sup supervisor exist?  I'm currently
    %% assuming that center wants to know when an interface dies so it
    %% can update its table, so it links itself to that child.  However,
    %% if the supervisor restarts the child (transient or permanent),
    %% how do we know who to link to?  That new child would have to tell
    %% us who it is.  Ick.  If the original child is temporary, then
    %% the supervisor will not restart it for us, so we'll have to tell
    %% the supervisor to start a new temporary child, but then why use
    %% the supervisor: just have center spawn_link it.  Perhaps the
    %% ethbr_iface_sup was to have a nice way to kill the pcap procs
    %% if the center died, but if they're linked directly to center,
    %% then they'll die, too?  Ugh.
    %%
    %% Well, leave ethbr_iface sup in for now.  It'll help make an
    %% "appmon" supervisor tree look prettier.  {shrug}
    %%
    case lists:keysearch(center_name, 1, ArgList) of
	{value, {_, Center}} ->
	    gen_server:cast(Center, ethbr_iface_sup_started),
	    %% Child_spec = {IdName, {M, F, A},
	    %%               Restart, Shutdown_time, Type, Modules_used}
	    {ok, {{one_for_one, 5, 3600}, [
					   %% Intentionally empty
					  ]}};
	_ ->
	    {error, {not_found, center_name}}
    end.

%%%----------------------------------------------------------------------
%%% Internal functions
%%%----------------------------------------------------------------------
