/*
** File    : libpcap_drv.c
** Summary : Erlang driver for the libpcap
** 
** NOTICE: This file was generated by the tools of the Erlang Driver
**         toolkit.  Do not edit this file by hand unless you know
**         what you're doing!
** 
** Copyright (c) 2002, Scott Lystig Fritchie.  All rights reserved.
** See the file "LICENSE" at the top of the source distribution for
** full license terms.
*/

/*
** XXX Before I forget yet again to write this down...
**
** ... the valmap ID assignment has a weakness that should be fixed
** sometime in the future.  The weakness is that reusing the array
** indexes could result in Erlang being able to access a later
** incarnation of a valmap table entry by simply remembering a
** previous valmap {valmap_blah, Integer} and resending it to the
** driver, hoping to get (un)lucky.  
**
** The solution would be to choose a valmap index from a larger, very
** unlikely to repeat set, and use a mapping data structure less naive
** than a dumb array.
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <assert.h>
#ifdef   DRIVER_USING_PTHREADS
#include <pthread.h>
#else    /* DRIVER_USING_PTHREADS */
#define  pthread_self()         0
#endif   /* DRIVER_USING_PTHREADS */

/* TODO: Add additional system & local header file #includes */

#include <erl_driver.h>
#include <erl_driver_tk.h>

/* <verbatim place="top_cpp_stuff"> */

#include <stdio.h>
#include <string.h>
#include <time.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include <pcap.h>
#include <libpcap_drv.h>     /* Gotta include this before my-libpcap.h */
#include <my-libpcap.h>

/*
** QQQ Alright, I haven't implemented _all_ of the libpcap API.
** For extra credit, list the API functions not yet supported.
*/


/* </verbatim --place="top_cpp_stuff"--> */

/* Last, but not least.... */
#include <libpcap_drv.h>

/* <verbatim place="top_other_stuff"> */

/* </verbatim --place="top_other_stuff"--> */

static ErlDrvTermData am_ok;
static ErlDrvTermData am_error;
static ErlDrvTermData am_badarg;
static ErlDrvTermData am_enomem;
static ErlDrvTermData am_unknown;
static ErlDrvTermData am_valmap_pcap_t;
static ErlDrvTermData am_valmap_bpf_program;

static int default_async_calls = 0;

/* This variable may be set only by pipe-main! */
int pipe_driver_p = 0;

/*
** TODO: Define any other Erlang terms this driver may return.
*/

/* Function prototypes */

ErlDrvEntry *driver_init(void *);       /* Do not change name! */
static int _init(void);
static ErlDrvData _start(ErlDrvPort port, char *command);
static void _stop(ErlDrvData drv_data);
static void _output(ErlDrvData drv_data, char *buf, int len);
static int _control(ErlDrvData drv_data, unsigned int command, char *buf, int len, char **rbuf, int rlen);
static void _outputv(ErlDrvData drv_data, ErlIOVec *ev);
static void _ready_async(ErlDrvData drv_data, ErlDrvThreadData thread_data);

void *sys_alloc(size_t);
void *sys_realloc(void *, size_t);
void sys_free(void *);

static void invoke__null(void *data);
static void invoke__pcap_open_live(void *data);
static void invoke__pcap_open_offline(void *data);
static void invoke__my_set_immediate(void *data);
static void invoke__my_set_see_sent(void *data);
static void invoke__pcap_setnonblock(void *data);
static void invoke__pcap_getnonblock(void *data);
static void invoke__pcap_findalldevs(void *data);
static void invoke__pcap_lookupdev(void *data);
static void invoke__pcap_lookupnet(void *data);
static void invoke__my_pcap_compile(void *data);
static void invoke__pcap_setfilter(void *data);
static void invoke__my_pcap_freecode(void *data);
static void invoke__pcap_next(void *data);
static void invoke__pcap_datalink(void *data);
static void invoke__pcap_snapshot(void *data);
static void invoke__pcap_is_swapped(void *data);
static void invoke__pcap_major_version(void *data);
static void invoke__pcap_minor_version(void *data);
static void invoke__pcap_stats(void *data);
static void invoke__pcap_fileno(void *data);
static void invoke__pcap_geterr(void *data);
static void invoke__pcap_close(void *data);

static int reply_xtra_pkt_len_in_pkthdr_else_nada(descriptor_t *, callstate_t *);
static int reply_xtra_ret_char_p_else_errbuf(descriptor_t *, callstate_t *);
static int reply_xtra_netp_mask_else_errbuf(descriptor_t *, callstate_t *);
static int reply_xtra_findalldevs(descriptor_t *, callstate_t *);
static int reply_xtra_pcap_stats(descriptor_t *, callstate_t *);

static int find_unused_pcap_t_index(descriptor_t *, unsigned long *);
static void cleanup_valmap_pcap_t_index(descriptor_t *, int, int);
static void cleanup_valmap_pcap_t_all(descriptor_t *, int);
static int find_unused_bpf_program_index(descriptor_t *, unsigned long *);
static void cleanup_valmap_bpf_program_index(descriptor_t *, int, int);
static void cleanup_valmap_bpf_program_all(descriptor_t *, int);

static int reply_ok(descriptor_t *desc);
static int reply_ok_num(descriptor_t *desc, unsigned long num);
static int reply_ok_binary(descriptor_t *desc, char *p, int, int);
static int reply_ok_valmap(descriptor_t *, ErlDrvTermData, unsigned long);
static int reply_error(descriptor_t *desc, int errnum);
static int reply_error_atom(descriptor_t *, ErlDrvTermData);
#if     0
static int reply_tag_ok(descriptor_t *desc, unsigned short tag);
static int reply_tag_error(descriptor_t *desc, unsigned short tag, int errnum);
#endif  /* 0 */

static ErlDrvEntry _driver_entry = {
    _init,                      /* init */
    _start,                     /* start */
    _stop,                      /* stop */
    _output,                    /* output */
    NULL,                       /* ready_input */
    NULL,                       /* ready_output */
    "libpcap_drv",              /* driver_name */
    NULL,                       /* finish */
    NULL,                       /* handle */
    _control,                   /* control */
    NULL,                       /* timeout */
    _outputv,                   /* outputv */
    _ready_async,               /* ready_async */
    NULL,                       /* flush */
    NULL                        /* call */
};

/*
** All dynamically-loadable driver libraries must contain a driver_init().
*/

ErlDrvEntry *
driver_init(void *handle)
{
    edtk_debug_flag = 1;
    _driver_entry.handle = handle;
    return &_driver_entry;
}

static int
_init(void)
{
    am_ok = driver_mk_atom("ok");
    am_error = driver_mk_atom("error");
    am_badarg = driver_mk_atom("badarg");
    am_enomem = driver_mk_atom("enomem");
    am_unknown = driver_mk_atom("unknown");
    am_valmap_pcap_t = driver_mk_atom("valmap_pcap_t");
    am_valmap_bpf_program = driver_mk_atom("valmap_bpf_program");

    /* TODO: Take care of other first-time initialization tasks */

    return 0;
}

static ErlDrvData
_start(ErlDrvPort port, char *args)
{
    descriptor_t        *desc;
    int                 i = 0;

    i = i;
    edtk_debug("%s: starting, port = %ld, args = 0x%lx, %s", __FUNCTION__,
               port, (unsigned long) args, args);

    if ((desc = (descriptor_t *) sys_alloc(sizeof(descriptor_t))) == NULL) {
        return ERL_DRV_ERROR_GENERAL;
    }
    memset(desc, 0, sizeof(descriptor_t));
    desc->port = port;
    desc->nextxid = 1;
    for (i = 0; i < 32; i++) {
        desc->valmap_pcap_t[i] = NULL;
        desc->valmap_pcap_t_flags[i] = 0;
    }
    for (i = 0; i < 32; i++) {
        desc->valmap_bpf_program[i] = NULL;
        desc->valmap_bpf_program_flags[i] = 0;
    }
    pthread_mutex_init(&desc->mutex, NULL);
    /* TODO: Finish initializing descriptor members */

    /* TODO: Take care of other port initialization tasks */

    return (ErlDrvData) desc;
}

static void
_stop(ErlDrvData drv_data)
{
    descriptor_t        *desc = (descriptor_t *) drv_data;
    int                 i = 0;
    ErlDrvPort          port;
    int                 still_in_use = 0;

    i = i;
    if (desc == NULL) {
        edtk_debug("%s: drv_data == NULL", __FUNCTION__);
        return;
    }
    edtk_debug("%s: port = %ld", __FUNCTION__, desc->port);

    for (i = 0; i < 32; i++) {
        /* Safe to look at flags here without getting mutex, right? */
        if (desc->valmap_pcap_t_flags[i] & VALMAP_INUSE) {
            edtk_debug("%s: desc->valmap_pcap_t_flags[%d] still in use!", __FUNCTION__, i);
            still_in_use++;
        } else
        if (desc->valmap_pcap_t[i] != NULL) {
            cleanup_valmap_pcap_t_index(desc, i, 1);
        }
    }
    for (i = 0; i < 32; i++) {
        /* Safe to look at flags here without getting mutex, right? */
        if (desc->valmap_bpf_program_flags[i] & VALMAP_INUSE) {
            edtk_debug("%s: desc->valmap_bpf_program_flags[%d] still in use!", __FUNCTION__, i);
            still_in_use++;
        } else
        if (desc->valmap_bpf_program[i] != NULL) {
            cleanup_valmap_bpf_program_index(desc, i, 1);
        }
    }

    if (! still_in_use) {
        port = desc->port;
        sys_free(desc);    
        edtk_debug("%s: port = %ld finished", __FUNCTION__, port);
    } else {
        /*
        ** XXX Oi, this is a sticky problem.  This port is being shut
        ** down, but we've still got some valmaps in use.  We have no
        ** way to tell the VM that we cannot be shut down safely right
        ** now, so what in the heck do we do?  It seems to me that we
        ** have two choices:
        **   1. Block the entire VM until all valmaps are idle, then
        **      clean them up and then return.
        **   2. Create a new thread that will take care of monitoring
        **      the valmaps & doing their final cleanup.  This stop
        **      function, executing in the main thread, can return
        **      to the VM right away.
        ** For the sake of simplicity, I'm going to implement #1
        ** for now.  This will get more complicated once EDTK supports
        ** private worker threads, so we won't bother getting fancy
        ** for now.
        */
        edtk_debug("%s: port = %ld has %d valmaps still in use!", __FUNCTION__, desc->port, still_in_use);
        sleep(1);
        _stop(drv_data);                /* Hope we don't run out of stack */
    }
}

static void
_output(ErlDrvData drv_data, char *buf, int len)
{
    /*
    ** Nobody should be calling this function because we've
    ** defined the "outputv" driver method, which BEAM will always
    ** used if it's available.  I just put a debug statement in
    ** here so that I might actually notice if the impossible ever
    ** happens....
    */
    edtk_debug("%s: XXX someone tried to call us, silly", __FUNCTION__);
}


static int
_control(ErlDrvData drv_data, unsigned int command,
                  char *buf, int len, char **rbuf, int rlen)
{
    char        *ret = *rbuf;
    int         retlen;

    /*
    ** Nobody should be calling this function either.
    */
    ret[0] = 1;
    retlen = 1;
    edtk_debug("%s: XXX someone tried to call us, silly", __FUNCTION__);
    return retlen;
}

static void
_outputv(ErlDrvData drv_data, ErlIOVec *ev)
{
    descriptor_t        *desc = (descriptor_t *) drv_data;
    unsigned char       cmd;
    int                 p = 0, q = 1;
    callstate_t         *c = NULL;
    int                 do_async_call = default_async_calls;
    unsigned long       binlen;
    int                 index;
    void                *tmp = NULL;

    binlen = binlen;
    index = index;
    tmp = tmp;
    if (desc == NULL || ev == NULL || ev->size < 1) {
        edtk_debug("%s: bad arg(s)", __FUNCTION__);
        return;
    }
    if (! EV_GET_CHAR(ev, &cmd, &p, &q)) {
        edtk_debug("%s: empty command", __FUNCTION__);
        reply_error(desc, EINVAL);
    }
    if ((c = sys_alloc(sizeof(callstate_t))) == NULL) {
        reply_error(desc, ENOMEM);
        return;
    }
    c->cmd = cmd;
    c->key = NULL;
    c->free = sys_free;
    c->xid = 0;         /* XXX unused right now */
    c->o.__expect = 1;  /* Default is that expectation is always met */
    c->desc = desc;
    c->refc_bincount = 0;

    edtk_debug("%s: my threadid = %lx, cmd = %d", __FUNCTION__, pthread_self(), cmd);
    switch (cmd) {
    case _DEBUG:
        EV_GET_UINT32(ev, &edtk_debug_flag, &p, &q);
        reply_ok_num(desc, edtk_debug_flag);    /* Immediate reply */
        sys_free(c);
        c = NULL;
        break;
    case _NULL:
        c->invoke = invoke__null;
        /* Do not ever do async call if we're in pipe driver mode */
        do_async_call =  pipe_driver_p ? 0 : do_async_call;
        break;
    case _PCAP_OPEN_LIVE:
        c->invoke = invoke__pcap_open_live;
        /* Do not ever do async call if we're in pipe driver mode */
        do_async_call =  pipe_driver_p ? 0 : do_async_call;
        EV_GET_UINT32(ev, &binlen, &p, &q);
        c->i.device = (char *) EV_GETPOS(ev, p, q);
        if (do_async_call) {
            if (c->refc_bincount == MAX_BINVS) {
                goto error;
            }
            ev->binv[q]->refc++;
            c->refc_bin[c->refc_bincount++] = ev->binv[q];
        }
        if (edtk_ev_forward_N(ev, binlen, &p, &q, 1) < 0) {
            goto error;
        }
        EV_GET_UINT32(ev, &c->i.snaplen, &p, &q);
        EV_GET_UINT32(ev, &c->i.promisc, &p, &q);
        EV_GET_UINT32(ev, &c->i.to_ms, &p, &q);
        /* <hack place="post-deserialize" type="verbatim"> */

        c->o.errbuf[0] = '\0';
 
        /* </hack --place="post-deserialize" type="verbatim"--> */
        break;
    case _PCAP_OPEN_OFFLINE:
        c->invoke = invoke__pcap_open_offline;
        /* Do not ever do async call if we're in pipe driver mode */
        do_async_call =  pipe_driver_p ? 0 : do_async_call;
        EV_GET_UINT32(ev, &binlen, &p, &q);
        c->i.fname = (char *) EV_GETPOS(ev, p, q);
        if (do_async_call) {
            if (c->refc_bincount == MAX_BINVS) {
                goto error;
            }
            ev->binv[q]->refc++;
            c->refc_bin[c->refc_bincount++] = ev->binv[q];
        }
        if (edtk_ev_forward_N(ev, binlen, &p, &q, 1) < 0) {
            goto error;
        }
        /* <hack place="post-deserialize" type="verbatim"> */

        c->o.errbuf[0] = '\0';
 
        /* </hack --place="post-deserialize" type="verbatim"--> */
        break;
    case _MY_SET_IMMEDIATE:
        c->invoke = invoke__my_set_immediate;
        /* Do not ever do async call if we're in pipe driver mode */
        do_async_call =  pipe_driver_p ? 0 : do_async_call;
        EV_GET_UINT32(ev, &index, &p, &q);
        if (index == -1 && 0 == 1) {
            edtk_debug("%s: valmap pcap_t index %d = default value 0x%lx", __FUNCTION__, index, NULL);
            c->i.p = NULL;
            c->i.__valmap_pcap_t_index = -1;
        } else if (desc->valmap_pcap_t[index] == NULL) {
            goto error;
        } else {
            edtk_debug("%s: valmap pcap_t index %d = 0x%lx", __FUNCTION__, index, desc->valmap_pcap_t[index]);
            c->i.p = desc->valmap_pcap_t[index];
            c->i.__valmap_pcap_t_index = index;
        }
        EV_GET_UINT32(ev, &c->i.immediate, &p, &q);
        /* Begin valmap flag adjustment */
        pthread_mutex_lock(&desc->mutex);
        if (c->i.__valmap_pcap_t_index >= 0) {
            edtk_debug("%s: desc->valmap_pcap_t_flags[%d] = %d\r\n", __FUNCTION__, c->i.__valmap_pcap_t_index, desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index]);
            assert(desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] == 0);
            desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] |= VALMAP_INUSE;
            edtk_debug("%s: INUSE set, desc->valmap_pcap_t_flags[%d] = %d\r\n", __FUNCTION__, c->i.__valmap_pcap_t_index, desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index]);
        }
        pthread_mutex_unlock(&desc->mutex);
        /* End valmap flag adjustment */
        break;
    case _MY_SET_SEE_SENT:
        c->invoke = invoke__my_set_see_sent;
        /* Do not ever do async call if we're in pipe driver mode */
        do_async_call =  pipe_driver_p ? 0 : do_async_call;
        EV_GET_UINT32(ev, &index, &p, &q);
        if (index == -1 && 0 == 1) {
            edtk_debug("%s: valmap pcap_t index %d = default value 0x%lx", __FUNCTION__, index, NULL);
            c->i.p = NULL;
            c->i.__valmap_pcap_t_index = -1;
        } else if (desc->valmap_pcap_t[index] == NULL) {
            goto error;
        } else {
            edtk_debug("%s: valmap pcap_t index %d = 0x%lx", __FUNCTION__, index, desc->valmap_pcap_t[index]);
            c->i.p = desc->valmap_pcap_t[index];
            c->i.__valmap_pcap_t_index = index;
        }
        EV_GET_UINT32(ev, &c->i.see_sent, &p, &q);
        /* Begin valmap flag adjustment */
        pthread_mutex_lock(&desc->mutex);
        if (c->i.__valmap_pcap_t_index >= 0) {
            edtk_debug("%s: desc->valmap_pcap_t_flags[%d] = %d\r\n", __FUNCTION__, c->i.__valmap_pcap_t_index, desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index]);
            assert(desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] == 0);
            desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] |= VALMAP_INUSE;
            edtk_debug("%s: INUSE set, desc->valmap_pcap_t_flags[%d] = %d\r\n", __FUNCTION__, c->i.__valmap_pcap_t_index, desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index]);
        }
        pthread_mutex_unlock(&desc->mutex);
        /* End valmap flag adjustment */
        break;
    case _PCAP_SETNONBLOCK:
        c->invoke = invoke__pcap_setnonblock;
        /* Do not ever do async call if we're in pipe driver mode */
        do_async_call =  pipe_driver_p ? 0 : do_async_call;
        EV_GET_UINT32(ev, &index, &p, &q);
        if (index == -1 && 0 == 1) {
            edtk_debug("%s: valmap pcap_t index %d = default value 0x%lx", __FUNCTION__, index, NULL);
            c->i.p = NULL;
            c->i.__valmap_pcap_t_index = -1;
        } else if (desc->valmap_pcap_t[index] == NULL) {
            goto error;
        } else {
            edtk_debug("%s: valmap pcap_t index %d = 0x%lx", __FUNCTION__, index, desc->valmap_pcap_t[index]);
            c->i.p = desc->valmap_pcap_t[index];
            c->i.__valmap_pcap_t_index = index;
        }
        EV_GET_UINT32(ev, &c->i.nonblock, &p, &q);
        /* Begin valmap flag adjustment */
        pthread_mutex_lock(&desc->mutex);
        if (c->i.__valmap_pcap_t_index >= 0) {
            edtk_debug("%s: desc->valmap_pcap_t_flags[%d] = %d\r\n", __FUNCTION__, c->i.__valmap_pcap_t_index, desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index]);
            assert(desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] == 0);
            desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] |= VALMAP_INUSE;
            edtk_debug("%s: INUSE set, desc->valmap_pcap_t_flags[%d] = %d\r\n", __FUNCTION__, c->i.__valmap_pcap_t_index, desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index]);
        }
        pthread_mutex_unlock(&desc->mutex);
        /* End valmap flag adjustment */
        break;
    case _PCAP_GETNONBLOCK:
        c->invoke = invoke__pcap_getnonblock;
        /* Do not ever do async call if we're in pipe driver mode */
        do_async_call =  pipe_driver_p ? 0 : do_async_call;
        EV_GET_UINT32(ev, &index, &p, &q);
        if (index == -1 && 0 == 1) {
            edtk_debug("%s: valmap pcap_t index %d = default value 0x%lx", __FUNCTION__, index, NULL);
            c->i.p = NULL;
            c->i.__valmap_pcap_t_index = -1;
        } else if (desc->valmap_pcap_t[index] == NULL) {
            goto error;
        } else {
            edtk_debug("%s: valmap pcap_t index %d = 0x%lx", __FUNCTION__, index, desc->valmap_pcap_t[index]);
            c->i.p = desc->valmap_pcap_t[index];
            c->i.__valmap_pcap_t_index = index;
        }
        /* Begin valmap flag adjustment */
        pthread_mutex_lock(&desc->mutex);
        if (c->i.__valmap_pcap_t_index >= 0) {
            edtk_debug("%s: desc->valmap_pcap_t_flags[%d] = %d\r\n", __FUNCTION__, c->i.__valmap_pcap_t_index, desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index]);
            assert(desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] == 0);
            desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] |= VALMAP_INUSE;
            edtk_debug("%s: INUSE set, desc->valmap_pcap_t_flags[%d] = %d\r\n", __FUNCTION__, c->i.__valmap_pcap_t_index, desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index]);
        }
        pthread_mutex_unlock(&desc->mutex);
        /* End valmap flag adjustment */
        break;
    case _PCAP_FINDALLDEVS:
        c->invoke = invoke__pcap_findalldevs;
        /* Do not ever do async call if we're in pipe driver mode */
        do_async_call =  pipe_driver_p ? 0 : do_async_call;
        /* <hack place="post-deserialize" type="verbatim"> */

        c->o.errbuf[0] = '\0';
        if (geteuid() != 0) {
                goto error;
        }
 
        /* </hack --place="post-deserialize" type="verbatim"--> */
        break;
    case _PCAP_LOOKUPDEV:
        c->invoke = invoke__pcap_lookupdev;
        /* Do not ever do async call if we're in pipe driver mode */
        do_async_call =  pipe_driver_p ? 0 : do_async_call;
        /* <hack place="post-deserialize" type="verbatim"> */

        c->o.errbuf[0] = '\0';
 
        /* </hack --place="post-deserialize" type="verbatim"--> */
        break;
    case _PCAP_LOOKUPNET:
        c->invoke = invoke__pcap_lookupnet;
        /* Do not ever do async call if we're in pipe driver mode */
        do_async_call =  pipe_driver_p ? 0 : do_async_call;
        EV_GET_UINT32(ev, &binlen, &p, &q);
        c->i.device = (char *) EV_GETPOS(ev, p, q);
        if (do_async_call) {
            if (c->refc_bincount == MAX_BINVS) {
                goto error;
            }
            ev->binv[q]->refc++;
            c->refc_bin[c->refc_bincount++] = ev->binv[q];
        }
        if (edtk_ev_forward_N(ev, binlen, &p, &q, 1) < 0) {
            goto error;
        }
        /* <hack place="post-deserialize" type="verbatim"> */

        c->o.errbuf[0] = '\0';
 
        /* </hack --place="post-deserialize" type="verbatim"--> */
        break;
    case _MY_PCAP_COMPILE:
        c->invoke = invoke__my_pcap_compile;
        do_async_call =  1;
        /* Do not ever do async call if we're in pipe driver mode */
        do_async_call =  pipe_driver_p ? 0 : do_async_call;
        EV_GET_UINT32(ev, &index, &p, &q);
        if (index == -1 && 0 == 1) {
            edtk_debug("%s: valmap pcap_t index %d = default value 0x%lx", __FUNCTION__, index, NULL);
            c->i.p = NULL;
            c->i.__valmap_pcap_t_index = -1;
        } else if (desc->valmap_pcap_t[index] == NULL) {
            goto error;
        } else {
            edtk_debug("%s: valmap pcap_t index %d = 0x%lx", __FUNCTION__, index, desc->valmap_pcap_t[index]);
            c->i.p = desc->valmap_pcap_t[index];
            c->i.__valmap_pcap_t_index = index;
        }
        EV_GET_UINT32(ev, &binlen, &p, &q);
        c->i.str = (char *) EV_GETPOS(ev, p, q);
        if (do_async_call) {
            if (c->refc_bincount == MAX_BINVS) {
                goto error;
            }
            ev->binv[q]->refc++;
            c->refc_bin[c->refc_bincount++] = ev->binv[q];
        }
        if (edtk_ev_forward_N(ev, binlen, &p, &q, 1) < 0) {
            goto error;
        }
        EV_GET_UINT32(ev, &c->i.optimize, &p, &q);
        EV_GET_UINT32(ev, &c->i.netmask, &p, &q);
        /* Begin valmap flag adjustment */
        pthread_mutex_lock(&desc->mutex);
        if (c->i.__valmap_pcap_t_index >= 0) {
            edtk_debug("%s: desc->valmap_pcap_t_flags[%d] = %d\r\n", __FUNCTION__, c->i.__valmap_pcap_t_index, desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index]);
            assert(desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] == 0);
            desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] |= VALMAP_INUSE;
            edtk_debug("%s: INUSE set, desc->valmap_pcap_t_flags[%d] = %d\r\n", __FUNCTION__, c->i.__valmap_pcap_t_index, desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index]);
        }
        pthread_mutex_unlock(&desc->mutex);
        /* End valmap flag adjustment */
        break;
    case _PCAP_SETFILTER:
        c->invoke = invoke__pcap_setfilter;
        /* Do not ever do async call if we're in pipe driver mode */
        do_async_call =  pipe_driver_p ? 0 : do_async_call;
        EV_GET_UINT32(ev, &index, &p, &q);
        if (index == -1 && 0 == 1) {
            edtk_debug("%s: valmap pcap_t index %d = default value 0x%lx", __FUNCTION__, index, NULL);
            c->i.p = NULL;
            c->i.__valmap_pcap_t_index = -1;
        } else if (desc->valmap_pcap_t[index] == NULL) {
            goto error;
        } else {
            edtk_debug("%s: valmap pcap_t index %d = 0x%lx", __FUNCTION__, index, desc->valmap_pcap_t[index]);
            c->i.p = desc->valmap_pcap_t[index];
            c->i.__valmap_pcap_t_index = index;
        }
        EV_GET_UINT32(ev, &index, &p, &q);
        if (index == -1 && 0 == 1) {
            edtk_debug("%s: valmap bpf_program index %d = default value 0x%lx", __FUNCTION__, index, NULL);
            c->i.fp = NULL;
            c->i.__valmap_bpf_program_index = -1;
        } else if (desc->valmap_bpf_program[index] == NULL) {
            goto error;
        } else {
            edtk_debug("%s: valmap bpf_program index %d = 0x%lx", __FUNCTION__, index, desc->valmap_bpf_program[index]);
            c->i.fp = desc->valmap_bpf_program[index];
            c->i.__valmap_bpf_program_index = index;
        }
        /* Begin valmap flag adjustment */
        pthread_mutex_lock(&desc->mutex);
        if (c->i.__valmap_pcap_t_index >= 0) {
            edtk_debug("%s: desc->valmap_pcap_t_flags[%d] = %d\r\n", __FUNCTION__, c->i.__valmap_pcap_t_index, desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index]);
            assert(desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] == 0);
            desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] |= VALMAP_INUSE;
            edtk_debug("%s: INUSE set, desc->valmap_pcap_t_flags[%d] = %d\r\n", __FUNCTION__, c->i.__valmap_pcap_t_index, desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index]);
        }
        if (c->i.__valmap_bpf_program_index >= 0) {
            edtk_debug("%s: desc->valmap_bpf_program_flags[%d] = %d\r\n", __FUNCTION__, c->i.__valmap_bpf_program_index, desc->valmap_bpf_program_flags[c->i.__valmap_bpf_program_index]);
            assert(desc->valmap_bpf_program_flags[c->i.__valmap_bpf_program_index] == 0);
            desc->valmap_bpf_program_flags[c->i.__valmap_bpf_program_index] |= VALMAP_INUSE;
            edtk_debug("%s: INUSE set, desc->valmap_bpf_program_flags[%d] = %d\r\n", __FUNCTION__, c->i.__valmap_bpf_program_index, desc->valmap_bpf_program_flags[c->i.__valmap_bpf_program_index]);
        }
        pthread_mutex_unlock(&desc->mutex);
        /* End valmap flag adjustment */
        break;
    case _MY_PCAP_FREECODE:
        c->invoke = invoke__my_pcap_freecode;
        /* Do not ever do async call if we're in pipe driver mode */
        do_async_call =  pipe_driver_p ? 0 : do_async_call;
        EV_GET_UINT32(ev, &index, &p, &q);
        if (index == -1 && 0 == 1) {
            edtk_debug("%s: valmap bpf_program index %d = default value 0x%lx", __FUNCTION__, index, NULL);
            c->i.fp = NULL;
            c->i.__valmap_bpf_program_index = -1;
        } else if (desc->valmap_bpf_program[index] == NULL) {
            goto error;
        } else {
            edtk_debug("%s: valmap bpf_program index %d = 0x%lx", __FUNCTION__, index, desc->valmap_bpf_program[index]);
            c->i.fp = desc->valmap_bpf_program[index];
            c->i.__valmap_bpf_program_index = index;
        }
        /* Begin valmap flag adjustment */
        pthread_mutex_lock(&desc->mutex);
        if (c->i.__valmap_bpf_program_index >= 0) {
            edtk_debug("%s: desc->valmap_bpf_program_flags[%d] = %d\r\n", __FUNCTION__, c->i.__valmap_bpf_program_index, desc->valmap_bpf_program_flags[c->i.__valmap_bpf_program_index]);
            assert(desc->valmap_bpf_program_flags[c->i.__valmap_bpf_program_index] == 0);
            desc->valmap_bpf_program_flags[c->i.__valmap_bpf_program_index] |= VALMAP_INUSE;
            edtk_debug("%s: INUSE set, desc->valmap_bpf_program_flags[%d] = %d\r\n", __FUNCTION__, c->i.__valmap_bpf_program_index, desc->valmap_bpf_program_flags[c->i.__valmap_bpf_program_index]);
        }
        pthread_mutex_unlock(&desc->mutex);
        /* End valmap flag adjustment */
        break;
    case _PCAP_NEXT:
        c->invoke = invoke__pcap_next;
        do_async_call =  1;
        /* Do not ever do async call if we're in pipe driver mode */
        do_async_call =  pipe_driver_p ? 0 : do_async_call;
        EV_GET_UINT32(ev, &index, &p, &q);
        if (index == -1 && 0 == 1) {
            edtk_debug("%s: valmap pcap_t index %d = default value 0x%lx", __FUNCTION__, index, NULL);
            c->i.p = NULL;
            c->i.__valmap_pcap_t_index = -1;
        } else if (desc->valmap_pcap_t[index] == NULL) {
            goto error;
        } else {
            edtk_debug("%s: valmap pcap_t index %d = 0x%lx", __FUNCTION__, index, desc->valmap_pcap_t[index]);
            c->i.p = desc->valmap_pcap_t[index];
            c->i.__valmap_pcap_t_index = index;
        }
        /* Begin valmap flag adjustment */
        pthread_mutex_lock(&desc->mutex);
        if (c->i.__valmap_pcap_t_index >= 0) {
            edtk_debug("%s: desc->valmap_pcap_t_flags[%d] = %d\r\n", __FUNCTION__, c->i.__valmap_pcap_t_index, desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index]);
            assert(desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] == 0);
            desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] |= VALMAP_INUSE;
            edtk_debug("%s: INUSE set, desc->valmap_pcap_t_flags[%d] = %d\r\n", __FUNCTION__, c->i.__valmap_pcap_t_index, desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index]);
        }
        pthread_mutex_unlock(&desc->mutex);
        /* End valmap flag adjustment */
        break;
    case _PCAP_DATALINK:
        c->invoke = invoke__pcap_datalink;
        /* Do not ever do async call if we're in pipe driver mode */
        do_async_call =  pipe_driver_p ? 0 : do_async_call;
        EV_GET_UINT32(ev, &index, &p, &q);
        if (index == -1 && 0 == 1) {
            edtk_debug("%s: valmap pcap_t index %d = default value 0x%lx", __FUNCTION__, index, NULL);
            c->i.p = NULL;
            c->i.__valmap_pcap_t_index = -1;
        } else if (desc->valmap_pcap_t[index] == NULL) {
            goto error;
        } else {
            edtk_debug("%s: valmap pcap_t index %d = 0x%lx", __FUNCTION__, index, desc->valmap_pcap_t[index]);
            c->i.p = desc->valmap_pcap_t[index];
            c->i.__valmap_pcap_t_index = index;
        }
        /* Begin valmap flag adjustment */
        pthread_mutex_lock(&desc->mutex);
        if (c->i.__valmap_pcap_t_index >= 0) {
            edtk_debug("%s: desc->valmap_pcap_t_flags[%d] = %d\r\n", __FUNCTION__, c->i.__valmap_pcap_t_index, desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index]);
            assert(desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] == 0);
            desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] |= VALMAP_INUSE;
            edtk_debug("%s: INUSE set, desc->valmap_pcap_t_flags[%d] = %d\r\n", __FUNCTION__, c->i.__valmap_pcap_t_index, desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index]);
        }
        pthread_mutex_unlock(&desc->mutex);
        /* End valmap flag adjustment */
        break;
    case _PCAP_SNAPSHOT:
        c->invoke = invoke__pcap_snapshot;
        /* Do not ever do async call if we're in pipe driver mode */
        do_async_call =  pipe_driver_p ? 0 : do_async_call;
        EV_GET_UINT32(ev, &index, &p, &q);
        if (index == -1 && 0 == 1) {
            edtk_debug("%s: valmap pcap_t index %d = default value 0x%lx", __FUNCTION__, index, NULL);
            c->i.p = NULL;
            c->i.__valmap_pcap_t_index = -1;
        } else if (desc->valmap_pcap_t[index] == NULL) {
            goto error;
        } else {
            edtk_debug("%s: valmap pcap_t index %d = 0x%lx", __FUNCTION__, index, desc->valmap_pcap_t[index]);
            c->i.p = desc->valmap_pcap_t[index];
            c->i.__valmap_pcap_t_index = index;
        }
        /* Begin valmap flag adjustment */
        pthread_mutex_lock(&desc->mutex);
        if (c->i.__valmap_pcap_t_index >= 0) {
            edtk_debug("%s: desc->valmap_pcap_t_flags[%d] = %d\r\n", __FUNCTION__, c->i.__valmap_pcap_t_index, desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index]);
            assert(desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] == 0);
            desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] |= VALMAP_INUSE;
            edtk_debug("%s: INUSE set, desc->valmap_pcap_t_flags[%d] = %d\r\n", __FUNCTION__, c->i.__valmap_pcap_t_index, desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index]);
        }
        pthread_mutex_unlock(&desc->mutex);
        /* End valmap flag adjustment */
        break;
    case _PCAP_IS_SWAPPED:
        c->invoke = invoke__pcap_is_swapped;
        /* Do not ever do async call if we're in pipe driver mode */
        do_async_call =  pipe_driver_p ? 0 : do_async_call;
        EV_GET_UINT32(ev, &index, &p, &q);
        if (index == -1 && 0 == 1) {
            edtk_debug("%s: valmap pcap_t index %d = default value 0x%lx", __FUNCTION__, index, NULL);
            c->i.p = NULL;
            c->i.__valmap_pcap_t_index = -1;
        } else if (desc->valmap_pcap_t[index] == NULL) {
            goto error;
        } else {
            edtk_debug("%s: valmap pcap_t index %d = 0x%lx", __FUNCTION__, index, desc->valmap_pcap_t[index]);
            c->i.p = desc->valmap_pcap_t[index];
            c->i.__valmap_pcap_t_index = index;
        }
        /* Begin valmap flag adjustment */
        pthread_mutex_lock(&desc->mutex);
        if (c->i.__valmap_pcap_t_index >= 0) {
            edtk_debug("%s: desc->valmap_pcap_t_flags[%d] = %d\r\n", __FUNCTION__, c->i.__valmap_pcap_t_index, desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index]);
            assert(desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] == 0);
            desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] |= VALMAP_INUSE;
            edtk_debug("%s: INUSE set, desc->valmap_pcap_t_flags[%d] = %d\r\n", __FUNCTION__, c->i.__valmap_pcap_t_index, desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index]);
        }
        pthread_mutex_unlock(&desc->mutex);
        /* End valmap flag adjustment */
        break;
    case _PCAP_MAJOR_VERSION:
        c->invoke = invoke__pcap_major_version;
        /* Do not ever do async call if we're in pipe driver mode */
        do_async_call =  pipe_driver_p ? 0 : do_async_call;
        EV_GET_UINT32(ev, &index, &p, &q);
        if (index == -1 && 0 == 1) {
            edtk_debug("%s: valmap pcap_t index %d = default value 0x%lx", __FUNCTION__, index, NULL);
            c->i.p = NULL;
            c->i.__valmap_pcap_t_index = -1;
        } else if (desc->valmap_pcap_t[index] == NULL) {
            goto error;
        } else {
            edtk_debug("%s: valmap pcap_t index %d = 0x%lx", __FUNCTION__, index, desc->valmap_pcap_t[index]);
            c->i.p = desc->valmap_pcap_t[index];
            c->i.__valmap_pcap_t_index = index;
        }
        /* Begin valmap flag adjustment */
        pthread_mutex_lock(&desc->mutex);
        if (c->i.__valmap_pcap_t_index >= 0) {
            edtk_debug("%s: desc->valmap_pcap_t_flags[%d] = %d\r\n", __FUNCTION__, c->i.__valmap_pcap_t_index, desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index]);
            assert(desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] == 0);
            desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] |= VALMAP_INUSE;
            edtk_debug("%s: INUSE set, desc->valmap_pcap_t_flags[%d] = %d\r\n", __FUNCTION__, c->i.__valmap_pcap_t_index, desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index]);
        }
        pthread_mutex_unlock(&desc->mutex);
        /* End valmap flag adjustment */
        break;
    case _PCAP_MINOR_VERSION:
        c->invoke = invoke__pcap_minor_version;
        /* Do not ever do async call if we're in pipe driver mode */
        do_async_call =  pipe_driver_p ? 0 : do_async_call;
        EV_GET_UINT32(ev, &index, &p, &q);
        if (index == -1 && 0 == 1) {
            edtk_debug("%s: valmap pcap_t index %d = default value 0x%lx", __FUNCTION__, index, NULL);
            c->i.p = NULL;
            c->i.__valmap_pcap_t_index = -1;
        } else if (desc->valmap_pcap_t[index] == NULL) {
            goto error;
        } else {
            edtk_debug("%s: valmap pcap_t index %d = 0x%lx", __FUNCTION__, index, desc->valmap_pcap_t[index]);
            c->i.p = desc->valmap_pcap_t[index];
            c->i.__valmap_pcap_t_index = index;
        }
        /* Begin valmap flag adjustment */
        pthread_mutex_lock(&desc->mutex);
        if (c->i.__valmap_pcap_t_index >= 0) {
            edtk_debug("%s: desc->valmap_pcap_t_flags[%d] = %d\r\n", __FUNCTION__, c->i.__valmap_pcap_t_index, desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index]);
            assert(desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] == 0);
            desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] |= VALMAP_INUSE;
            edtk_debug("%s: INUSE set, desc->valmap_pcap_t_flags[%d] = %d\r\n", __FUNCTION__, c->i.__valmap_pcap_t_index, desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index]);
        }
        pthread_mutex_unlock(&desc->mutex);
        /* End valmap flag adjustment */
        break;
    case _PCAP_STATS:
        c->invoke = invoke__pcap_stats;
        /* Do not ever do async call if we're in pipe driver mode */
        do_async_call =  pipe_driver_p ? 0 : do_async_call;
        EV_GET_UINT32(ev, &index, &p, &q);
        if (index == -1 && 0 == 1) {
            edtk_debug("%s: valmap pcap_t index %d = default value 0x%lx", __FUNCTION__, index, NULL);
            c->i.p = NULL;
            c->i.__valmap_pcap_t_index = -1;
        } else if (desc->valmap_pcap_t[index] == NULL) {
            goto error;
        } else {
            edtk_debug("%s: valmap pcap_t index %d = 0x%lx", __FUNCTION__, index, desc->valmap_pcap_t[index]);
            c->i.p = desc->valmap_pcap_t[index];
            c->i.__valmap_pcap_t_index = index;
        }
        /* Begin valmap flag adjustment */
        pthread_mutex_lock(&desc->mutex);
        if (c->i.__valmap_pcap_t_index >= 0) {
            edtk_debug("%s: desc->valmap_pcap_t_flags[%d] = %d\r\n", __FUNCTION__, c->i.__valmap_pcap_t_index, desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index]);
            assert(desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] == 0);
            desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] |= VALMAP_INUSE;
            edtk_debug("%s: INUSE set, desc->valmap_pcap_t_flags[%d] = %d\r\n", __FUNCTION__, c->i.__valmap_pcap_t_index, desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index]);
        }
        pthread_mutex_unlock(&desc->mutex);
        /* End valmap flag adjustment */
        break;
    case _PCAP_FILENO:
        c->invoke = invoke__pcap_fileno;
        /* Do not ever do async call if we're in pipe driver mode */
        do_async_call =  pipe_driver_p ? 0 : do_async_call;
        EV_GET_UINT32(ev, &index, &p, &q);
        if (index == -1 && 0 == 1) {
            edtk_debug("%s: valmap pcap_t index %d = default value 0x%lx", __FUNCTION__, index, NULL);
            c->i.p = NULL;
            c->i.__valmap_pcap_t_index = -1;
        } else if (desc->valmap_pcap_t[index] == NULL) {
            goto error;
        } else {
            edtk_debug("%s: valmap pcap_t index %d = 0x%lx", __FUNCTION__, index, desc->valmap_pcap_t[index]);
            c->i.p = desc->valmap_pcap_t[index];
            c->i.__valmap_pcap_t_index = index;
        }
        /* Begin valmap flag adjustment */
        pthread_mutex_lock(&desc->mutex);
        if (c->i.__valmap_pcap_t_index >= 0) {
            edtk_debug("%s: desc->valmap_pcap_t_flags[%d] = %d\r\n", __FUNCTION__, c->i.__valmap_pcap_t_index, desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index]);
            assert(desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] == 0);
            desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] |= VALMAP_INUSE;
            edtk_debug("%s: INUSE set, desc->valmap_pcap_t_flags[%d] = %d\r\n", __FUNCTION__, c->i.__valmap_pcap_t_index, desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index]);
        }
        pthread_mutex_unlock(&desc->mutex);
        /* End valmap flag adjustment */
        break;
    case _PCAP_GETERR:
        c->invoke = invoke__pcap_geterr;
        /* Do not ever do async call if we're in pipe driver mode */
        do_async_call =  pipe_driver_p ? 0 : do_async_call;
        EV_GET_UINT32(ev, &index, &p, &q);
        if (index == -1 && 0 == 1) {
            edtk_debug("%s: valmap pcap_t index %d = default value 0x%lx", __FUNCTION__, index, NULL);
            c->i.p = NULL;
            c->i.__valmap_pcap_t_index = -1;
        } else if (desc->valmap_pcap_t[index] == NULL) {
            goto error;
        } else {
            edtk_debug("%s: valmap pcap_t index %d = 0x%lx", __FUNCTION__, index, desc->valmap_pcap_t[index]);
            c->i.p = desc->valmap_pcap_t[index];
            c->i.__valmap_pcap_t_index = index;
        }
        /* Begin valmap flag adjustment */
        pthread_mutex_lock(&desc->mutex);
        if (c->i.__valmap_pcap_t_index >= 0) {
            edtk_debug("%s: desc->valmap_pcap_t_flags[%d] = %d\r\n", __FUNCTION__, c->i.__valmap_pcap_t_index, desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index]);
            assert(desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] == 0);
            desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] |= VALMAP_INUSE;
            edtk_debug("%s: INUSE set, desc->valmap_pcap_t_flags[%d] = %d\r\n", __FUNCTION__, c->i.__valmap_pcap_t_index, desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index]);
        }
        pthread_mutex_unlock(&desc->mutex);
        /* End valmap flag adjustment */
        break;
    case _PCAP_CLOSE:
        c->invoke = invoke__pcap_close;
        /* Do not ever do async call if we're in pipe driver mode */
        do_async_call =  pipe_driver_p ? 0 : do_async_call;
        EV_GET_UINT32(ev, &index, &p, &q);
        if (index == -1 && 0 == 1) {
            edtk_debug("%s: valmap pcap_t index %d = default value 0x%lx", __FUNCTION__, index, NULL);
            c->i.p = NULL;
            c->i.__valmap_pcap_t_index = -1;
        } else if (desc->valmap_pcap_t[index] == NULL) {
            goto error;
        } else {
            edtk_debug("%s: valmap pcap_t index %d = 0x%lx", __FUNCTION__, index, desc->valmap_pcap_t[index]);
            c->i.p = desc->valmap_pcap_t[index];
            c->i.__valmap_pcap_t_index = index;
        }
        /* Begin valmap flag adjustment */
        pthread_mutex_lock(&desc->mutex);
        if (c->i.__valmap_pcap_t_index >= 0) {
            edtk_debug("%s: desc->valmap_pcap_t_flags[%d] = %d\r\n", __FUNCTION__, c->i.__valmap_pcap_t_index, desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index]);
            assert(desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] == 0);
            desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] |= VALMAP_INUSE;
            edtk_debug("%s: INUSE set, desc->valmap_pcap_t_flags[%d] = %d\r\n", __FUNCTION__, c->i.__valmap_pcap_t_index, desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index]);
        }
        pthread_mutex_unlock(&desc->mutex);
        /* End valmap flag adjustment */
        break;
     default:
        edtk_debug("%s: invalid command %d", __FUNCTION__, cmd);
        goto error;
        break;  
    }
    if (c != NULL) {
        if (do_async_call) {
            driver_async(desc->port, c->key, c->invoke, c, c->free);
        } else {
            /*
            ** Execute the bottom half right away, then send the result.
            */
            (*(c->invoke))((void *) c);
            _ready_async((ErlDrvData) desc, (ErlDrvThreadData) c);
            /* 
            ** c is already freed for us by _ready_async()
            */
        }
    }
    return;

  error:
    if (c != NULL) {
        sys_free(c);
    }
    reply_error_atom(desc, am_badarg);
}

static void
_ready_async(ErlDrvData drv_data, ErlDrvThreadData thread_data)
{
    descriptor_t        *desc = (descriptor_t *) drv_data;
    callstate_t *c = (callstate_t *) thread_data;
    int                 bytes, offset, i;
    char                *p = NULL;
    unsigned long       index = 0;

    p = p;
    bytes = bytes;
    offset = offset;
    i = i;
    index = index;
    edtk_debug("%s: cmd = %d", __FUNCTION__, c->cmd);
    if (c == NULL) {
        edtk_debug("%s: c == NULL", __FUNCTION__);
        return;
    }
    switch (c->cmd) {
    case _NULL:
        reply_ok(desc);
        break;
    case _PCAP_OPEN_LIVE:
        if (! c->o.__expect) {
            reply_error(desc, c->o.__expect_errval);
            break;
        }
        if (find_unused_pcap_t_index(desc, &index) < 0) {
            reply_error(desc, ENOMEM);
        } else {
            desc->valmap_pcap_t[index] = c->o.ret_pcap_t;
            reply_ok_valmap(desc, am_valmap_pcap_t, index);
        }
        break;
    case _PCAP_OPEN_OFFLINE:
        if (! c->o.__expect) {
            reply_error(desc, c->o.__expect_errval);
            break;
        }
        if (find_unused_pcap_t_index(desc, &index) < 0) {
            reply_error(desc, ENOMEM);
        } else {
            desc->valmap_pcap_t[index] = c->o.ret_pcap_t;
            reply_ok_valmap(desc, am_valmap_pcap_t, index);
        }
        break;
    case _MY_SET_IMMEDIATE:
        if (c->o.__expect) {
            reply_ok_num(desc, c->o.ret_int);
        } else {
            reply_error(desc, c->o.__expect_errval);
        }
        break;
    case _MY_SET_SEE_SENT:
        if (c->o.__expect) {
            reply_ok_num(desc, c->o.ret_int);
        } else {
            reply_error(desc, c->o.__expect_errval);
        }
        break;
    case _PCAP_SETNONBLOCK:
        if (c->o.__expect) {
            reply_ok_num(desc, c->o.ret_int);
        } else {
            reply_error(desc, c->o.__expect_errval);
        }
        break;
    case _PCAP_GETNONBLOCK:
        if (c->o.__expect) {
            reply_ok_num(desc, c->o.ret_int);
        } else {
            reply_error(desc, c->o.__expect_errval);
        }
        break;
    case _PCAP_FINDALLDEVS:
        reply_xtra_findalldevs(desc, c);
        /* <hack place="ready_async-post-reply" type="verbatim"> */

        if (c->o.alldevsp != NULL) {
                pcap_freealldevs(c->o.alldevsp);
        }
 
        /* </hack --place="ready_async-post-reply" type="verbatim"--> */
        break;
    case _PCAP_LOOKUPDEV:
        reply_xtra_ret_char_p_else_errbuf(desc, c);
        break;
    case _PCAP_LOOKUPNET:
        reply_xtra_netp_mask_else_errbuf(desc, c);
        break;
    case _MY_PCAP_COMPILE:
        if (! c->o.__expect) {
            reply_error(desc, c->o.__expect_errval);
            break;
        }
        if (find_unused_bpf_program_index(desc, &index) < 0) {
            reply_error(desc, ENOMEM);
        } else {
            desc->valmap_bpf_program[index] = c->o.ret_bpf_program;
            reply_ok_valmap(desc, am_valmap_bpf_program, index);
        }
        break;
    case _PCAP_SETFILTER:
        if (c->o.__expect) {
            reply_ok_num(desc, c->o.ret_int);
        } else {
            reply_error(desc, c->o.__expect_errval);
        }
        break;
    case _MY_PCAP_FREECODE:
        cleanup_valmap_bpf_program_index(desc, c->i.__valmap_bpf_program_index, 0);
        reply_ok(desc);
        break;
    case _PCAP_NEXT:
        reply_xtra_pkt_len_in_pkthdr_else_nada(desc, c);
        break;
    case _PCAP_DATALINK:
        if (c->o.__expect) {
            reply_ok_num(desc, c->o.ret_int);
        } else {
            reply_error(desc, c->o.__expect_errval);
        }
        break;
    case _PCAP_SNAPSHOT:
        if (c->o.__expect) {
            reply_ok_num(desc, c->o.ret_int);
        } else {
            reply_error(desc, c->o.__expect_errval);
        }
        break;
    case _PCAP_IS_SWAPPED:
        if (c->o.__expect) {
            reply_ok_num(desc, c->o.ret_int);
        } else {
            reply_error(desc, c->o.__expect_errval);
        }
        break;
    case _PCAP_MAJOR_VERSION:
        if (c->o.__expect) {
            reply_ok_num(desc, c->o.ret_int);
        } else {
            reply_error(desc, c->o.__expect_errval);
        }
        break;
    case _PCAP_MINOR_VERSION:
        if (c->o.__expect) {
            reply_ok_num(desc, c->o.ret_int);
        } else {
            reply_error(desc, c->o.__expect_errval);
        }
        break;
    case _PCAP_STATS:
        reply_xtra_pcap_stats(desc, c);
        break;
    case _PCAP_FILENO:
        if (c->o.__expect) {
            reply_ok_num(desc, c->o.ret_int);
        } else {
            reply_error(desc, c->o.__expect_errval);
        }
        break;
    case _PCAP_GETERR:
        bytes = strlen(c->o.ret_char_p);
        offset = 0;
        if ((p = edtk_driver_alloc_wrapper(bytes)) == NULL) {
            reply_error(desc, ENOMEM);
        } else {
            memcpy(p, c->o.ret_char_p, bytes);
            reply_ok_binary(desc, p, 0, bytes);
        }
        break;
    case _PCAP_CLOSE:
        cleanup_valmap_pcap_t_index(desc, c->i.__valmap_pcap_t_index, 0);
        reply_ok(desc);
        break;
    default:
        edtk_debug("%s: bogus command, should never happen", __FUNCTION__);
        break;
    }    
    for (i = 0; i < c->refc_bincount; i++) {
        driver_free_binary(c->refc_bin[i]);
    }
    sys_free(c);
}

static void
invoke__null(void *data)
{
    callstate_t *c = (callstate_t *) data;

    c = c;
    edtk_debug("%s: threadid = %lx", __FUNCTION__, pthread_self());
    null(
                        );
    edtk_debug("%s: threadid = %lx done", __FUNCTION__, pthread_self());
}

static void
invoke__pcap_open_live(void *data)
{
    callstate_t *c = (callstate_t *) data;

    c = c;
    edtk_debug("%s: threadid = %lx", __FUNCTION__, pthread_self());
    c->o.ret_pcap_t = pcap_open_live(
                          c->i.device, 
                          c->i.snaplen, 
                          c->i.promisc, 
                          c->i.to_ms, 
                          c->o.errbuf
                        );
    if (c->o.ret_pcap_t != NULL) {
        c->o.__expect = 1;
    } else {
        c->o.__expect = 0;
        c->o.__expect_errval = errno;
        /* Danger!  Do not put debugging statement before saving error val! */
        edtk_debug("%s: threadid = %lx expectation failed!", __FUNCTION__, pthread_self());
    }
    edtk_debug("%s: threadid = %lx done", __FUNCTION__, pthread_self());
}

static void
invoke__pcap_open_offline(void *data)
{
    callstate_t *c = (callstate_t *) data;

    c = c;
    edtk_debug("%s: threadid = %lx", __FUNCTION__, pthread_self());
    c->o.ret_pcap_t = pcap_open_offline(
                          c->i.fname, 
                          c->o.errbuf
                        );
    if (c->o.ret_pcap_t != 0) {
        c->o.__expect = 1;
    } else {
        c->o.__expect = 0;
        c->o.__expect_errval = errno;
        /* Danger!  Do not put debugging statement before saving error val! */
        edtk_debug("%s: threadid = %lx expectation failed!", __FUNCTION__, pthread_self());
    }
    edtk_debug("%s: threadid = %lx done", __FUNCTION__, pthread_self());
}

static void
invoke__my_set_immediate(void *data)
{
    callstate_t *c = (callstate_t *) data;

    c = c;
    edtk_debug("%s: threadid = %lx", __FUNCTION__, pthread_self());
    c->o.ret_int = my_set_immediate(
                          c->i.p, 
                          c->i.immediate
                        );
    if (c->o.ret_int != -1) {
        c->o.__expect = 1;
    } else {
        c->o.__expect = 0;
        c->o.__expect_errval = errno;
        /* Danger!  Do not put debugging statement before saving error val! */
        edtk_debug("%s: threadid = %lx expectation failed!", __FUNCTION__, pthread_self());
    }
    /* Begin valmap flag adjustment */
    pthread_mutex_lock(&c->desc->mutex);
    if (c->i.__valmap_pcap_t_index >= 0) {
        c->desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] &= (~VALMAP_INUSE);
        edtk_debug("%s: INUSE cleared, c->desc->valmap_pcap_t_flags[%d] = %d\r\n", __FUNCTION__, c->i.__valmap_pcap_t_index, c->desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index]);
        if (c->desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] & VALMAP_DELAYED_CLEANUP) {
            cleanup_valmap_pcap_t_index(c->desc, c->i.__valmap_pcap_t_index, 1);
        }
    }
    pthread_mutex_unlock(&c->desc->mutex);
    /* End valmap flag adjustment */
    edtk_debug("%s: threadid = %lx done", __FUNCTION__, pthread_self());
}

static void
invoke__my_set_see_sent(void *data)
{
    callstate_t *c = (callstate_t *) data;

    c = c;
    edtk_debug("%s: threadid = %lx", __FUNCTION__, pthread_self());
    c->o.ret_int = my_set_see_sent(
                          c->i.p, 
                          c->i.see_sent
                        );
    if (c->o.ret_int != -1) {
        c->o.__expect = 1;
    } else {
        c->o.__expect = 0;
        c->o.__expect_errval = errno;
        /* Danger!  Do not put debugging statement before saving error val! */
        edtk_debug("%s: threadid = %lx expectation failed!", __FUNCTION__, pthread_self());
    }
    /* Begin valmap flag adjustment */
    pthread_mutex_lock(&c->desc->mutex);
    if (c->i.__valmap_pcap_t_index >= 0) {
        c->desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] &= (~VALMAP_INUSE);
        edtk_debug("%s: INUSE cleared, c->desc->valmap_pcap_t_flags[%d] = %d\r\n", __FUNCTION__, c->i.__valmap_pcap_t_index, c->desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index]);
        if (c->desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] & VALMAP_DELAYED_CLEANUP) {
            cleanup_valmap_pcap_t_index(c->desc, c->i.__valmap_pcap_t_index, 1);
        }
    }
    pthread_mutex_unlock(&c->desc->mutex);
    /* End valmap flag adjustment */
    edtk_debug("%s: threadid = %lx done", __FUNCTION__, pthread_self());
}

static void
invoke__pcap_setnonblock(void *data)
{
    callstate_t *c = (callstate_t *) data;

    c = c;
    edtk_debug("%s: threadid = %lx", __FUNCTION__, pthread_self());
    c->o.ret_int = pcap_setnonblock(
                          c->i.p, 
                          c->i.nonblock, 
                          c->o.errbuf
                        );
    if (c->o.ret_int != -1) {
        c->o.__expect = 1;
    } else {
        c->o.__expect = 0;
        c->o.__expect_errval = errno;
        /* Danger!  Do not put debugging statement before saving error val! */
        edtk_debug("%s: threadid = %lx expectation failed!", __FUNCTION__, pthread_self());
    }
    /* Begin valmap flag adjustment */
    pthread_mutex_lock(&c->desc->mutex);
    if (c->i.__valmap_pcap_t_index >= 0) {
        c->desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] &= (~VALMAP_INUSE);
        edtk_debug("%s: INUSE cleared, c->desc->valmap_pcap_t_flags[%d] = %d\r\n", __FUNCTION__, c->i.__valmap_pcap_t_index, c->desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index]);
        if (c->desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] & VALMAP_DELAYED_CLEANUP) {
            cleanup_valmap_pcap_t_index(c->desc, c->i.__valmap_pcap_t_index, 1);
        }
    }
    pthread_mutex_unlock(&c->desc->mutex);
    /* End valmap flag adjustment */
    edtk_debug("%s: threadid = %lx done", __FUNCTION__, pthread_self());
}

static void
invoke__pcap_getnonblock(void *data)
{
    callstate_t *c = (callstate_t *) data;

    c = c;
    edtk_debug("%s: threadid = %lx", __FUNCTION__, pthread_self());
    c->o.ret_int = pcap_getnonblock(
                          c->i.p, 
                          c->o.errbuf
                        );
    if (c->o.ret_int != -1) {
        c->o.__expect = 1;
    } else {
        c->o.__expect = 0;
        c->o.__expect_errval = errno;
        /* Danger!  Do not put debugging statement before saving error val! */
        edtk_debug("%s: threadid = %lx expectation failed!", __FUNCTION__, pthread_self());
    }
    /* Begin valmap flag adjustment */
    pthread_mutex_lock(&c->desc->mutex);
    if (c->i.__valmap_pcap_t_index >= 0) {
        c->desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] &= (~VALMAP_INUSE);
        edtk_debug("%s: INUSE cleared, c->desc->valmap_pcap_t_flags[%d] = %d\r\n", __FUNCTION__, c->i.__valmap_pcap_t_index, c->desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index]);
        if (c->desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] & VALMAP_DELAYED_CLEANUP) {
            cleanup_valmap_pcap_t_index(c->desc, c->i.__valmap_pcap_t_index, 1);
        }
    }
    pthread_mutex_unlock(&c->desc->mutex);
    /* End valmap flag adjustment */
    edtk_debug("%s: threadid = %lx done", __FUNCTION__, pthread_self());
}

static void
invoke__pcap_findalldevs(void *data)
{
    callstate_t *c = (callstate_t *) data;

    c = c;
    edtk_debug("%s: threadid = %lx", __FUNCTION__, pthread_self());
    c->o.ret_int = pcap_findalldevs(
                         & c->o.alldevsp, 
                          c->o.errbuf
                        );
    if (c->o.ret_int != -1) {
        c->o.__expect = 1;
    } else {
        c->o.__expect = 0;
        c->o.__expect_errval = -1;
        /* Danger!  Do not put debugging statement before saving error val! */
        edtk_debug("%s: threadid = %lx expectation failed!", __FUNCTION__, pthread_self());
    }
    edtk_debug("%s: threadid = %lx done", __FUNCTION__, pthread_self());
}

static void
invoke__pcap_lookupdev(void *data)
{
    callstate_t *c = (callstate_t *) data;

    c = c;
    edtk_debug("%s: threadid = %lx", __FUNCTION__, pthread_self());
    c->o.ret_char_p = pcap_lookupdev(
                          c->o.errbuf
                        );
    if (c->o.ret_char_p != NULL) {
        c->o.__expect = 1;
    } else {
        c->o.__expect = 0;
        c->o.__expect_errval = -1;
        /* Danger!  Do not put debugging statement before saving error val! */
        edtk_debug("%s: threadid = %lx expectation failed!", __FUNCTION__, pthread_self());
    }
    edtk_debug("%s: threadid = %lx done", __FUNCTION__, pthread_self());
}

static void
invoke__pcap_lookupnet(void *data)
{
    callstate_t *c = (callstate_t *) data;

    c = c;
    edtk_debug("%s: threadid = %lx", __FUNCTION__, pthread_self());
    c->o.ret_int = pcap_lookupnet(
                          c->i.device, 
                         & c->o.netp, 
                         & c->o.mask, 
                          c->o.errbuf
                        );
    if (c->o.ret_int != -1) {
        c->o.__expect = 1;
    } else {
        c->o.__expect = 0;
        c->o.__expect_errval = -1;
        /* Danger!  Do not put debugging statement before saving error val! */
        edtk_debug("%s: threadid = %lx expectation failed!", __FUNCTION__, pthread_self());
    }
    edtk_debug("%s: threadid = %lx done", __FUNCTION__, pthread_self());
}

static void
invoke__my_pcap_compile(void *data)
{
    callstate_t *c = (callstate_t *) data;

    c = c;
    edtk_debug("%s: threadid = %lx", __FUNCTION__, pthread_self());
    c->o.ret_bpf_program = my_pcap_compile(
                          c->i.p, 
                          c->i.str, 
                          c->i.optimize, 
                          c->i.netmask
                        );
    if (c->o.ret_bpf_program != NULL) {
        c->o.__expect = 1;
    } else {
        c->o.__expect = 0;
        c->o.__expect_errval = errno;
        /* Danger!  Do not put debugging statement before saving error val! */
        edtk_debug("%s: threadid = %lx expectation failed!", __FUNCTION__, pthread_self());
    }
    /* Begin valmap flag adjustment */
    pthread_mutex_lock(&c->desc->mutex);
    if (c->i.__valmap_pcap_t_index >= 0) {
        c->desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] &= (~VALMAP_INUSE);
        edtk_debug("%s: INUSE cleared, c->desc->valmap_pcap_t_flags[%d] = %d\r\n", __FUNCTION__, c->i.__valmap_pcap_t_index, c->desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index]);
        if (c->desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] & VALMAP_DELAYED_CLEANUP) {
            cleanup_valmap_pcap_t_index(c->desc, c->i.__valmap_pcap_t_index, 1);
        }
    }
    pthread_mutex_unlock(&c->desc->mutex);
    /* End valmap flag adjustment */
    edtk_debug("%s: threadid = %lx done", __FUNCTION__, pthread_self());
}

static void
invoke__pcap_setfilter(void *data)
{
    callstate_t *c = (callstate_t *) data;

    c = c;
    edtk_debug("%s: threadid = %lx", __FUNCTION__, pthread_self());
    c->o.ret_int = pcap_setfilter(
                          c->i.p, 
                          c->i.fp
                        );
    if (c->o.ret_int != -1) {
        c->o.__expect = 1;
    } else {
        c->o.__expect = 0;
        c->o.__expect_errval = errno;
        /* Danger!  Do not put debugging statement before saving error val! */
        edtk_debug("%s: threadid = %lx expectation failed!", __FUNCTION__, pthread_self());
    }
    /* Begin valmap flag adjustment */
    pthread_mutex_lock(&c->desc->mutex);
    if (c->i.__valmap_pcap_t_index >= 0) {
        c->desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] &= (~VALMAP_INUSE);
        edtk_debug("%s: INUSE cleared, c->desc->valmap_pcap_t_flags[%d] = %d\r\n", __FUNCTION__, c->i.__valmap_pcap_t_index, c->desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index]);
        if (c->desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] & VALMAP_DELAYED_CLEANUP) {
            cleanup_valmap_pcap_t_index(c->desc, c->i.__valmap_pcap_t_index, 1);
        }
    }
    if (c->i.__valmap_bpf_program_index >= 0) {
        c->desc->valmap_bpf_program_flags[c->i.__valmap_bpf_program_index] &= (~VALMAP_INUSE);
        edtk_debug("%s: INUSE cleared, c->desc->valmap_bpf_program_flags[%d] = %d\r\n", __FUNCTION__, c->i.__valmap_bpf_program_index, c->desc->valmap_bpf_program_flags[c->i.__valmap_bpf_program_index]);
        if (c->desc->valmap_bpf_program_flags[c->i.__valmap_bpf_program_index] & VALMAP_DELAYED_CLEANUP) {
            cleanup_valmap_bpf_program_index(c->desc, c->i.__valmap_bpf_program_index, 1);
        }
    }
    pthread_mutex_unlock(&c->desc->mutex);
    /* End valmap flag adjustment */
    edtk_debug("%s: threadid = %lx done", __FUNCTION__, pthread_self());
}

static void
invoke__my_pcap_freecode(void *data)
{
    callstate_t *c = (callstate_t *) data;

    c = c;
    edtk_debug("%s: threadid = %lx", __FUNCTION__, pthread_self());
    my_pcap_freecode(
                          c->i.fp
                        );
    /* Begin valmap flag adjustment */
    pthread_mutex_lock(&c->desc->mutex);
    if (c->i.__valmap_bpf_program_index >= 0) {
        c->desc->valmap_bpf_program_flags[c->i.__valmap_bpf_program_index] &= (~VALMAP_INUSE);
        edtk_debug("%s: INUSE cleared, c->desc->valmap_bpf_program_flags[%d] = %d\r\n", __FUNCTION__, c->i.__valmap_bpf_program_index, c->desc->valmap_bpf_program_flags[c->i.__valmap_bpf_program_index]);
        if (c->desc->valmap_bpf_program_flags[c->i.__valmap_bpf_program_index] & VALMAP_DELAYED_CLEANUP) {
            cleanup_valmap_bpf_program_index(c->desc, c->i.__valmap_bpf_program_index, 1);
        }
    }
    pthread_mutex_unlock(&c->desc->mutex);
    /* End valmap flag adjustment */
    edtk_debug("%s: threadid = %lx done", __FUNCTION__, pthread_self());
}

static void
invoke__pcap_next(void *data)
{
    callstate_t *c = (callstate_t *) data;

    c = c;
    edtk_debug("%s: threadid = %lx", __FUNCTION__, pthread_self());
    c->o.ret_u_char = pcap_next(
                          c->i.p, 
                         & c->o.h
                        );
    if (c->o.ret_u_char != NULL) {
        c->o.__expect = 1;
    } else {
        c->o.__expect = 0;
        c->o.__expect_errval = -1;
        /* Danger!  Do not put debugging statement before saving error val! */
        edtk_debug("%s: threadid = %lx expectation failed!", __FUNCTION__, pthread_self());
    }
    /* Begin valmap flag adjustment */
    pthread_mutex_lock(&c->desc->mutex);
    if (c->i.__valmap_pcap_t_index >= 0) {
        c->desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] &= (~VALMAP_INUSE);
        edtk_debug("%s: INUSE cleared, c->desc->valmap_pcap_t_flags[%d] = %d\r\n", __FUNCTION__, c->i.__valmap_pcap_t_index, c->desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index]);
        if (c->desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] & VALMAP_DELAYED_CLEANUP) {
            cleanup_valmap_pcap_t_index(c->desc, c->i.__valmap_pcap_t_index, 1);
        }
    }
    pthread_mutex_unlock(&c->desc->mutex);
    /* End valmap flag adjustment */
    edtk_debug("%s: threadid = %lx done", __FUNCTION__, pthread_self());
}

static void
invoke__pcap_datalink(void *data)
{
    callstate_t *c = (callstate_t *) data;

    c = c;
    edtk_debug("%s: threadid = %lx", __FUNCTION__, pthread_self());
    c->o.ret_int = pcap_datalink(
                          c->i.p
                        );
    if (c->o.ret_int != -1) {
        c->o.__expect = 1;
    } else {
        c->o.__expect = 0;
        c->o.__expect_errval = errno;
        /* Danger!  Do not put debugging statement before saving error val! */
        edtk_debug("%s: threadid = %lx expectation failed!", __FUNCTION__, pthread_self());
    }
    /* Begin valmap flag adjustment */
    pthread_mutex_lock(&c->desc->mutex);
    if (c->i.__valmap_pcap_t_index >= 0) {
        c->desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] &= (~VALMAP_INUSE);
        edtk_debug("%s: INUSE cleared, c->desc->valmap_pcap_t_flags[%d] = %d\r\n", __FUNCTION__, c->i.__valmap_pcap_t_index, c->desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index]);
        if (c->desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] & VALMAP_DELAYED_CLEANUP) {
            cleanup_valmap_pcap_t_index(c->desc, c->i.__valmap_pcap_t_index, 1);
        }
    }
    pthread_mutex_unlock(&c->desc->mutex);
    /* End valmap flag adjustment */
    edtk_debug("%s: threadid = %lx done", __FUNCTION__, pthread_self());
}

static void
invoke__pcap_snapshot(void *data)
{
    callstate_t *c = (callstate_t *) data;

    c = c;
    edtk_debug("%s: threadid = %lx", __FUNCTION__, pthread_self());
    c->o.ret_int = pcap_snapshot(
                          c->i.p
                        );
    if (c->o.ret_int != -1) {
        c->o.__expect = 1;
    } else {
        c->o.__expect = 0;
        c->o.__expect_errval = errno;
        /* Danger!  Do not put debugging statement before saving error val! */
        edtk_debug("%s: threadid = %lx expectation failed!", __FUNCTION__, pthread_self());
    }
    /* Begin valmap flag adjustment */
    pthread_mutex_lock(&c->desc->mutex);
    if (c->i.__valmap_pcap_t_index >= 0) {
        c->desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] &= (~VALMAP_INUSE);
        edtk_debug("%s: INUSE cleared, c->desc->valmap_pcap_t_flags[%d] = %d\r\n", __FUNCTION__, c->i.__valmap_pcap_t_index, c->desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index]);
        if (c->desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] & VALMAP_DELAYED_CLEANUP) {
            cleanup_valmap_pcap_t_index(c->desc, c->i.__valmap_pcap_t_index, 1);
        }
    }
    pthread_mutex_unlock(&c->desc->mutex);
    /* End valmap flag adjustment */
    edtk_debug("%s: threadid = %lx done", __FUNCTION__, pthread_self());
}

static void
invoke__pcap_is_swapped(void *data)
{
    callstate_t *c = (callstate_t *) data;

    c = c;
    edtk_debug("%s: threadid = %lx", __FUNCTION__, pthread_self());
    c->o.ret_int = pcap_is_swapped(
                          c->i.p
                        );
    if (c->o.ret_int != -1) {
        c->o.__expect = 1;
    } else {
        c->o.__expect = 0;
        c->o.__expect_errval = errno;
        /* Danger!  Do not put debugging statement before saving error val! */
        edtk_debug("%s: threadid = %lx expectation failed!", __FUNCTION__, pthread_self());
    }
    /* Begin valmap flag adjustment */
    pthread_mutex_lock(&c->desc->mutex);
    if (c->i.__valmap_pcap_t_index >= 0) {
        c->desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] &= (~VALMAP_INUSE);
        edtk_debug("%s: INUSE cleared, c->desc->valmap_pcap_t_flags[%d] = %d\r\n", __FUNCTION__, c->i.__valmap_pcap_t_index, c->desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index]);
        if (c->desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] & VALMAP_DELAYED_CLEANUP) {
            cleanup_valmap_pcap_t_index(c->desc, c->i.__valmap_pcap_t_index, 1);
        }
    }
    pthread_mutex_unlock(&c->desc->mutex);
    /* End valmap flag adjustment */
    edtk_debug("%s: threadid = %lx done", __FUNCTION__, pthread_self());
}

static void
invoke__pcap_major_version(void *data)
{
    callstate_t *c = (callstate_t *) data;

    c = c;
    edtk_debug("%s: threadid = %lx", __FUNCTION__, pthread_self());
    c->o.ret_int = pcap_major_version(
                          c->i.p
                        );
    if (c->o.ret_int != -1) {
        c->o.__expect = 1;
    } else {
        c->o.__expect = 0;
        c->o.__expect_errval = errno;
        /* Danger!  Do not put debugging statement before saving error val! */
        edtk_debug("%s: threadid = %lx expectation failed!", __FUNCTION__, pthread_self());
    }
    /* Begin valmap flag adjustment */
    pthread_mutex_lock(&c->desc->mutex);
    if (c->i.__valmap_pcap_t_index >= 0) {
        c->desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] &= (~VALMAP_INUSE);
        edtk_debug("%s: INUSE cleared, c->desc->valmap_pcap_t_flags[%d] = %d\r\n", __FUNCTION__, c->i.__valmap_pcap_t_index, c->desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index]);
        if (c->desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] & VALMAP_DELAYED_CLEANUP) {
            cleanup_valmap_pcap_t_index(c->desc, c->i.__valmap_pcap_t_index, 1);
        }
    }
    pthread_mutex_unlock(&c->desc->mutex);
    /* End valmap flag adjustment */
    edtk_debug("%s: threadid = %lx done", __FUNCTION__, pthread_self());
}

static void
invoke__pcap_minor_version(void *data)
{
    callstate_t *c = (callstate_t *) data;

    c = c;
    edtk_debug("%s: threadid = %lx", __FUNCTION__, pthread_self());
    c->o.ret_int = pcap_minor_version(
                          c->i.p
                        );
    if (c->o.ret_int != -1) {
        c->o.__expect = 1;
    } else {
        c->o.__expect = 0;
        c->o.__expect_errval = errno;
        /* Danger!  Do not put debugging statement before saving error val! */
        edtk_debug("%s: threadid = %lx expectation failed!", __FUNCTION__, pthread_self());
    }
    /* Begin valmap flag adjustment */
    pthread_mutex_lock(&c->desc->mutex);
    if (c->i.__valmap_pcap_t_index >= 0) {
        c->desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] &= (~VALMAP_INUSE);
        edtk_debug("%s: INUSE cleared, c->desc->valmap_pcap_t_flags[%d] = %d\r\n", __FUNCTION__, c->i.__valmap_pcap_t_index, c->desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index]);
        if (c->desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] & VALMAP_DELAYED_CLEANUP) {
            cleanup_valmap_pcap_t_index(c->desc, c->i.__valmap_pcap_t_index, 1);
        }
    }
    pthread_mutex_unlock(&c->desc->mutex);
    /* End valmap flag adjustment */
    edtk_debug("%s: threadid = %lx done", __FUNCTION__, pthread_self());
}

static void
invoke__pcap_stats(void *data)
{
    callstate_t *c = (callstate_t *) data;

    c = c;
    edtk_debug("%s: threadid = %lx", __FUNCTION__, pthread_self());
    c->o.ret_int = pcap_stats(
                          c->i.p, 
                         & c->o.ps
                        );
    if (c->o.ret_int != -1) {
        c->o.__expect = 1;
    } else {
        c->o.__expect = 0;
        c->o.__expect_errval = errno;
        /* Danger!  Do not put debugging statement before saving error val! */
        edtk_debug("%s: threadid = %lx expectation failed!", __FUNCTION__, pthread_self());
    }
    /* Begin valmap flag adjustment */
    pthread_mutex_lock(&c->desc->mutex);
    if (c->i.__valmap_pcap_t_index >= 0) {
        c->desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] &= (~VALMAP_INUSE);
        edtk_debug("%s: INUSE cleared, c->desc->valmap_pcap_t_flags[%d] = %d\r\n", __FUNCTION__, c->i.__valmap_pcap_t_index, c->desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index]);
        if (c->desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] & VALMAP_DELAYED_CLEANUP) {
            cleanup_valmap_pcap_t_index(c->desc, c->i.__valmap_pcap_t_index, 1);
        }
    }
    pthread_mutex_unlock(&c->desc->mutex);
    /* End valmap flag adjustment */
    edtk_debug("%s: threadid = %lx done", __FUNCTION__, pthread_self());
}

static void
invoke__pcap_fileno(void *data)
{
    callstate_t *c = (callstate_t *) data;

    c = c;
    edtk_debug("%s: threadid = %lx", __FUNCTION__, pthread_self());
    c->o.ret_int = pcap_fileno(
                          c->i.p
                        );
    if (c->o.ret_int != -1) {
        c->o.__expect = 1;
    } else {
        c->o.__expect = 0;
        c->o.__expect_errval = errno;
        /* Danger!  Do not put debugging statement before saving error val! */
        edtk_debug("%s: threadid = %lx expectation failed!", __FUNCTION__, pthread_self());
    }
    /* Begin valmap flag adjustment */
    pthread_mutex_lock(&c->desc->mutex);
    if (c->i.__valmap_pcap_t_index >= 0) {
        c->desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] &= (~VALMAP_INUSE);
        edtk_debug("%s: INUSE cleared, c->desc->valmap_pcap_t_flags[%d] = %d\r\n", __FUNCTION__, c->i.__valmap_pcap_t_index, c->desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index]);
        if (c->desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] & VALMAP_DELAYED_CLEANUP) {
            cleanup_valmap_pcap_t_index(c->desc, c->i.__valmap_pcap_t_index, 1);
        }
    }
    pthread_mutex_unlock(&c->desc->mutex);
    /* End valmap flag adjustment */
    edtk_debug("%s: threadid = %lx done", __FUNCTION__, pthread_self());
}

static void
invoke__pcap_geterr(void *data)
{
    callstate_t *c = (callstate_t *) data;

    c = c;
    edtk_debug("%s: threadid = %lx", __FUNCTION__, pthread_self());
    c->o.ret_char_p = pcap_geterr(
                          c->i.p
                        );
    if (c->o.ret_char_p != NULL) {
        c->o.__expect = 1;
    } else {
        c->o.__expect = 0;
        c->o.__expect_errval = errno;
        /* Danger!  Do not put debugging statement before saving error val! */
        edtk_debug("%s: threadid = %lx expectation failed!", __FUNCTION__, pthread_self());
    }
    /* Begin valmap flag adjustment */
    pthread_mutex_lock(&c->desc->mutex);
    if (c->i.__valmap_pcap_t_index >= 0) {
        c->desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] &= (~VALMAP_INUSE);
        edtk_debug("%s: INUSE cleared, c->desc->valmap_pcap_t_flags[%d] = %d\r\n", __FUNCTION__, c->i.__valmap_pcap_t_index, c->desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index]);
        if (c->desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] & VALMAP_DELAYED_CLEANUP) {
            cleanup_valmap_pcap_t_index(c->desc, c->i.__valmap_pcap_t_index, 1);
        }
    }
    pthread_mutex_unlock(&c->desc->mutex);
    /* End valmap flag adjustment */
    edtk_debug("%s: threadid = %lx done", __FUNCTION__, pthread_self());
}

static void
invoke__pcap_close(void *data)
{
    callstate_t *c = (callstate_t *) data;

    c = c;
    edtk_debug("%s: threadid = %lx", __FUNCTION__, pthread_self());
    pcap_close(
                          c->i.p
                        );
    /* Begin valmap flag adjustment */
    pthread_mutex_lock(&c->desc->mutex);
    if (c->i.__valmap_pcap_t_index >= 0) {
        c->desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] &= (~VALMAP_INUSE);
        edtk_debug("%s: INUSE cleared, c->desc->valmap_pcap_t_flags[%d] = %d\r\n", __FUNCTION__, c->i.__valmap_pcap_t_index, c->desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index]);
        if (c->desc->valmap_pcap_t_flags[c->i.__valmap_pcap_t_index] & VALMAP_DELAYED_CLEANUP) {
            cleanup_valmap_pcap_t_index(c->desc, c->i.__valmap_pcap_t_index, 1);
        }
    }
    pthread_mutex_unlock(&c->desc->mutex);
    /* End valmap flag adjustment */
    edtk_debug("%s: threadid = %lx done", __FUNCTION__, pthread_self());
}


static int
reply_ok(descriptor_t *desc)
{
    ErlDrvTermData      msg[24];
    int                 i = 0;
    int                 res;

    i = LOAD_PORT(msg, i, driver_mk_port(desc->port));
    i = LOAD_ATOM(msg, i, am_ok);
    i = LOAD_TUPLE(msg, i, 2);
    edtk_debug("reply_ok: i = %d", i);
    res = driver_output_term(desc->port, msg, i);
    edtk_debug("reply_ok: res = %d", res);
    return res;
}

static int
reply_ok_num(descriptor_t *desc, unsigned long num)
{
    ErlDrvTermData      msg[24];
    int                 i = 0;
    int                 res;

    i = LOAD_PORT(msg, i, driver_mk_port(desc->port));
    i = LOAD_ATOM(msg, i, am_ok);
    i = LOAD_INT(msg, i, num);
    i = LOAD_TUPLE(msg, i, 3);
    edtk_debug("%s: i = %d, num = %lu", __FUNCTION__, i, num);
    res = driver_output_term(desc->port, msg, i);
    edtk_debug("%s: res = %d", __FUNCTION__, res);
    return res;
}

static int
reply_ok_binary(descriptor_t *desc, char *ptr, int beg_offset, int length)
{
    ErlDrvTermData      msg[24];
    int                 i = 0;
    int                 res;

    i = LOAD_PORT(msg, i, driver_mk_port(desc->port));
    i = LOAD_ATOM(msg, i, am_ok);
    i = LOAD_BINARY(msg, i, edtk_alloced_ptr2ErlDrvBinary(ptr),
                    beg_offset, length);
    i = LOAD_TUPLE(msg, i, 3);
    edtk_debug("%s: i = %d, ptr = 0x%lx, start = %d, end = %d",
    __FUNCTION__, i, ptr, beg_offset, length);
    res = driver_output_term(desc->port, msg, i);
    /* driver_output_term() incrs refc, and we're done, so decr refc */
    /*
    ** We _know_ that "ptr" points to memory allocated by
    ** edtk_driver_alloc_wrapper(), so edtk_alloced_ptr2ErlDrvBinary()
    ** is safe in this case.  If it weren't safe, then the binary
    ** must be returned by an xtra_return, which means we
    ** reply_ok_binary()) are never called!
    */
    driver_free_binary(edtk_alloced_ptr2ErlDrvBinary(ptr));
    edtk_debug("%s: res = %d", __FUNCTION__, res);
    return res;
}

static int
reply_ok_valmap(descriptor_t *desc, ErlDrvTermData valmap_atom,
                unsigned long valmap_index)
{
    ErlDrvTermData      msg[15];
    int                 i = 0;
    int                 res;

    i = LOAD_PORT(msg, i, driver_mk_port(desc->port));
    i = LOAD_ATOM(msg, i, am_ok);
    i = LOAD_ATOM(msg, i, valmap_atom);
    i = LOAD_INT(msg, i, valmap_index);
    i = LOAD_TUPLE(msg, i, 2);
    i = LOAD_TUPLE(msg, i, 3);
    edtk_debug("reply_ok_valmap: i = %d", i);
    res = driver_output_term(desc->port, msg, i);
    edtk_debug("reply_ok_valmap: res = %d", res);
    return res;
}

static int
reply_error(descriptor_t *desc, int errnum)
{
    ErlDrvTermData      msg[24];
    int                 i = 0;
    int                 res;

    edtk_debug("reply_error: errnum = %d", errnum);
    i = LOAD_PORT(msg, i, driver_mk_port(desc->port));
    i = LOAD_ATOM(msg, i, am_error);
    i = LOAD_INT(msg, i, errnum);
    i = LOAD_TUPLE(msg, i, 3);
    edtk_debug("reply_error: i = %d", i);
    res = driver_output_term(desc->port, msg, i);
    edtk_debug("reply_error: res = %d", res);
    return res;
}

static int
reply_error_atom(descriptor_t *desc, ErlDrvTermData atom)
{
    ErlDrvTermData      msg[24];
    int                 i = 0;
    int                 res;

    i = LOAD_PORT(msg, i, driver_mk_port(desc->port));
    i = LOAD_ATOM(msg, i, am_error);
    i = LOAD_ATOM(msg, i, atom);
    i = LOAD_TUPLE(msg, i, 3);
    edtk_debug("%s: i = %d", __FUNCTION__, i);
    res = driver_output_term(desc->port, msg, i);
    edtk_debug("%s: res = %d", __FUNCTION__, res);
    return res;
}

#if     0       /* XXX These are unused right now */
static int
reply_tag_ok(descriptor_t *desc, unsigned short tag)
{
    ErlDrvTermData      msg[24];
    int                 i = 0;
    int                 res;

    i = LOAD_PORT(msg, i, driver_mk_port(desc->port));
    i = LOAD_INT(msg, i, tag);
    i = LOAD_ATOM(msg, i, am_ok);
    i = LOAD_TUPLE(msg, i, 3);
    edtk_debug("reply_ok: i = %d", i);
    res = driver_output_term(desc->port, msg, i);
    edtk_debug("reply_ok: res = %d", res);
    return res;
}

static int
reply_tag_error(descriptor_t *desc, unsigned short tag, int errnum)
{
    ErlDrvTermData      msg[24];
    int                 i = 0;
    int                 res;

    edtk_debug("reply_error: errnum = %d", errnum);
    i = LOAD_PORT(msg, i, driver_mk_port(desc->port));
    i = LOAD_INT(msg, i, tag);
    i = LOAD_INT(msg, i, errnum);
    i = LOAD_ATOM(msg, i, am_error);
    i = LOAD_TUPLE(msg, i, 4);
    edtk_debug("reply_error: i = %d", i);
    res = driver_output_term(desc->port, msg, i);
    edtk_debug("reply_error: res = %d", res);
    return res;
}
#endif  /* 0 */


static int
reply_xtra_pkt_len_in_pkthdr_else_nada(descriptor_t *desc, callstate_t *c)
{
    ErlDrvTermData      msg[MAX_RETURN_TERMS];
    int                 msgcount = 0;
    int                 res;
    int                 members = 0;
    char                *tmp = NULL;
    int                 i;
    ErlDrvBinary        *tmpbin = NULL;

    tmp = tmp; tmpbin = tmpbin;
    desc->num_tofree = 0;

    msgcount = LOAD_PORT(msg, msgcount, driver_mk_port(desc->port));
    members = 2;        /* members will be 2 very shortly: Port & ok|error */
    if (c->o.__expect) {
        msgcount = LOAD_ATOM(msg, msgcount, am_ok);
        {
            int members = 0;
             {
                 int members = 0;

                 msgcount = LOAD_INT(msg, msgcount, c->o.h.ts.tv_sec);
                 members++;
                 msgcount = LOAD_INT(msg, msgcount, c->o.h.ts.tv_usec);
                 members++;
                 msgcount = LOAD_TUPLE(msg, msgcount, members);
             }
             members++;
             if ((tmp = edtk_driver_alloc_wrapper(c->o.h.caplen)) == NULL) {
              return reply_error(desc, ENOMEM);
             }
             memcpy(tmp,  c->o.ret_u_char, c->o.h.caplen);
             tmpbin = edtk_alloced_ptr2ErlDrvBinary(tmp);
             msgcount = LOAD_BINARY(msg, msgcount, tmpbin, 0, c->o.h.caplen);
             /* driver_output_term() incrs refc, and we're done, so decr refc LATER */
             desc->tofree[desc->num_tofree++] = tmpbin;
             members++;
            msgcount = LOAD_TUPLE(msg, msgcount, members);
        }
    } else {
        msgcount = LOAD_ATOM(msg, msgcount, am_error);
        {
            int members = 0;
            msgcount = LOAD_TUPLE(msg, msgcount, members);
        }
    }
    msgcount = LOAD_TUPLE(msg, msgcount, 3);
    edtk_debug("reply_xtra_pkt_len_in_pkthdr_else_nada: i = %d", msgcount);
    res = driver_output_term(desc->port, msg, msgcount);
    edtk_debug("reply_xtra_pkt_len_in_pkthdr_else_nada: res = %d", res);
    if (res < 0) {
        fprintf(stderr, "\r\n\r\nreply_xtra_pkt_len_in_pkthdr_else_nada: driver_output_term() failed!  This should never happen!\r\n\r\n");
    }

    for (i = 0; i < desc->num_tofree; i++) {
        driver_free_binary(desc->tofree[i]);
    }
    return res;
}

static int
reply_xtra_ret_char_p_else_errbuf(descriptor_t *desc, callstate_t *c)
{
    ErlDrvTermData      msg[MAX_RETURN_TERMS];
    int                 msgcount = 0;
    int                 res;
    int                 members = 0;
    char                *tmp = NULL;
    int                 i;
    ErlDrvBinary        *tmpbin = NULL;

    tmp = tmp; tmpbin = tmpbin;
    desc->num_tofree = 0;

    msgcount = LOAD_PORT(msg, msgcount, driver_mk_port(desc->port));
    members = 2;        /* members will be 2 very shortly: Port & ok|error */
    if (c->o.__expect) {
        msgcount = LOAD_ATOM(msg, msgcount, am_ok);
        {
            int members = 0;
             msgcount = LOAD_STRING(msg, msgcount, c->o.ret_char_p, strlen(c->o.ret_char_p));
             members++;
            msgcount = LOAD_TUPLE(msg, msgcount, members);
        }
    } else {
        msgcount = LOAD_ATOM(msg, msgcount, am_error);
        {
            int members = 0;
             msgcount = LOAD_STRING(msg, msgcount, c->o.errbuf, strlen(c->o.errbuf));
             members++;
            msgcount = LOAD_TUPLE(msg, msgcount, members);
        }
    }
    msgcount = LOAD_TUPLE(msg, msgcount, 3);
    edtk_debug("reply_xtra_ret_char_p_else_errbuf: i = %d", msgcount);
    res = driver_output_term(desc->port, msg, msgcount);
    edtk_debug("reply_xtra_ret_char_p_else_errbuf: res = %d", res);
    if (res < 0) {
        fprintf(stderr, "\r\n\r\nreply_xtra_ret_char_p_else_errbuf: driver_output_term() failed!  This should never happen!\r\n\r\n");
    }

    for (i = 0; i < desc->num_tofree; i++) {
        driver_free_binary(desc->tofree[i]);
    }
    return res;
}

static int
reply_xtra_netp_mask_else_errbuf(descriptor_t *desc, callstate_t *c)
{
    ErlDrvTermData      msg[MAX_RETURN_TERMS];
    int                 msgcount = 0;
    int                 res;
    int                 members = 0;
    char                *tmp = NULL;
    int                 i;
    ErlDrvBinary        *tmpbin = NULL;

    tmp = tmp; tmpbin = tmpbin;
    desc->num_tofree = 0;

    msgcount = LOAD_PORT(msg, msgcount, driver_mk_port(desc->port));
    members = 2;        /* members will be 2 very shortly: Port & ok|error */
    if (c->o.__expect) {
        msgcount = LOAD_ATOM(msg, msgcount, am_ok);
        {
            int members = 0;
             msgcount = LOAD_INT(msg, msgcount, c->o.netp);
             members++;
             msgcount = LOAD_INT(msg, msgcount, c->o.mask);
             members++;
            msgcount = LOAD_TUPLE(msg, msgcount, members);
        }
    } else {
        msgcount = LOAD_ATOM(msg, msgcount, am_error);
        {
            int members = 0;
             msgcount = LOAD_STRING(msg, msgcount, c->o.errbuf, strlen(c->o.errbuf));
             members++;
            msgcount = LOAD_TUPLE(msg, msgcount, members);
        }
    }
    msgcount = LOAD_TUPLE(msg, msgcount, 3);
    edtk_debug("reply_xtra_netp_mask_else_errbuf: i = %d", msgcount);
    res = driver_output_term(desc->port, msg, msgcount);
    edtk_debug("reply_xtra_netp_mask_else_errbuf: res = %d", res);
    if (res < 0) {
        fprintf(stderr, "\r\n\r\nreply_xtra_netp_mask_else_errbuf: driver_output_term() failed!  This should never happen!\r\n\r\n");
    }

    for (i = 0; i < desc->num_tofree; i++) {
        driver_free_binary(desc->tofree[i]);
    }
    return res;
}

static int
reply_xtra_findalldevs(descriptor_t *desc, callstate_t *c)
{
    ErlDrvTermData      msg[MAX_RETURN_TERMS];
    int                 msgcount = 0;
    int                 res;
    int                 members = 0;
    char                *tmp = NULL;
    int                 i;
    ErlDrvBinary        *tmpbin = NULL;

    tmp = tmp; tmpbin = tmpbin;
    desc->num_tofree = 0;

    msgcount = LOAD_PORT(msg, msgcount, driver_mk_port(desc->port));
    members = 2;        /* members will be 2 very shortly: Port & ok|error */
    if (c->o.__expect) {
        msgcount = LOAD_ATOM(msg, msgcount, am_ok);
        {
            int members = 0;
             {
                 int members = 0;
                 
                 make_findalldevs_list(desc, c, msg, &members, &msgcount);
                 msgcount = LOAD_LIST(msg, msgcount, members);
             }
             members++;
            msgcount = LOAD_TUPLE(msg, msgcount, members);
        }
    } else {
        msgcount = LOAD_ATOM(msg, msgcount, am_error);
        {
            int members = 0;
             msgcount = LOAD_STRING(msg, msgcount, c->o.errbuf, strlen(c->o.errbuf));
             members++;
            msgcount = LOAD_TUPLE(msg, msgcount, members);
        }
    }
    msgcount = LOAD_TUPLE(msg, msgcount, 3);
    edtk_debug("reply_xtra_findalldevs: i = %d", msgcount);
    res = driver_output_term(desc->port, msg, msgcount);
    edtk_debug("reply_xtra_findalldevs: res = %d", res);
    if (res < 0) {
        fprintf(stderr, "\r\n\r\nreply_xtra_findalldevs: driver_output_term() failed!   This should never happen!\r\n\r\n");
    }

    for (i = 0; i < desc->num_tofree; i++) {
        driver_free_binary(desc->tofree[i]);
    }
    return res;
}

static int
reply_xtra_pcap_stats(descriptor_t *desc, callstate_t *c)
{
    ErlDrvTermData      msg[MAX_RETURN_TERMS];
    int                 msgcount = 0;
    int                 res;
    int                 members = 0;
    char                *tmp = NULL;
    int                 i;
    ErlDrvBinary        *tmpbin = NULL;

    tmp = tmp; tmpbin = tmpbin;
    desc->num_tofree = 0;

    msgcount = LOAD_PORT(msg, msgcount, driver_mk_port(desc->port));
    members = 2;        /* members will be 2 very shortly: Port & ok|error */
    if (c->o.__expect) {
        msgcount = LOAD_ATOM(msg, msgcount, am_ok);
        {
            int members = 0;
             msgcount = LOAD_INT(msg, msgcount, c->o.ps.ps_recv);
             members++;
             msgcount = LOAD_INT(msg, msgcount, c->o.ps.ps_drop);
             members++;
            msgcount = LOAD_TUPLE(msg, msgcount, members);
        }
    } else {
        msgcount = LOAD_ATOM(msg, msgcount, am_error);
        {
            int members = 0;
            msgcount = LOAD_TUPLE(msg, msgcount, members);
        }
    }
    msgcount = LOAD_TUPLE(msg, msgcount, 3);
    edtk_debug("reply_xtra_pcap_stats: i = %d", msgcount);
    res = driver_output_term(desc->port, msg, msgcount);
    edtk_debug("reply_xtra_pcap_stats: res = %d", res);
    if (res < 0) {
        fprintf(stderr, "\r\n\r\nreply_xtra_pcap_stats: driver_output_term() failed!    This should never happen!\r\n\r\n");
    }

    for (i = 0; i < desc->num_tofree; i++) {
        driver_free_binary(desc->tofree[i]);
    }
    return res;
}


static int
find_unused_pcap_t_index(descriptor_t *desc, unsigned long *index_p)
{
    int i;

    for (i = 0; i < 32; i++) {
        if (desc->valmap_pcap_t[i] == NULL) {
            *index_p = i;
            return 0;
        }
    }
    return -1;
}

static void
cleanup_valmap_pcap_t_index(descriptor_t *desc, int i, int do_cleanup_func)
{
    edtk_debug("%s: i = %d, do_cleanup_func = %d", __FUNCTION__, i, do_cleanup_func);
    pthread_mutex_lock(&desc->mutex);
    if (desc->valmap_pcap_t_flags[i] & VALMAP_INUSE) {
        edtk_debug("%s: i = %d, in use, setting VALMAP_DELAYED_CLEANUP flag", __FUNCTION__, i);
        desc->valmap_pcap_t_flags[i] |= VALMAP_DELAYED_CLEANUP;
        pthread_mutex_unlock(&desc->mutex);
        return;
    }
    desc->valmap_pcap_t_flags[i] = 0;
    pthread_mutex_unlock(&desc->mutex);
    if (do_cleanup_func) {
        edtk_debug("%s: calling func pcap_close", __FUNCTION__);
        pcap_close(desc->valmap_pcap_t[i]);
    }
    desc->valmap_pcap_t[i] = NULL;
}

static void
cleanup_valmap_pcap_t_all(descriptor_t *desc, int do_cleanup_func)
{
    int i;

    edtk_debug("%s: do_cleanup_func = %d", __FUNCTION__, do_cleanup_func);
    for (i = 0; i < 32; i++) {
        cleanup_valmap_pcap_t_index(desc, i, do_cleanup_func);
    }
}

static int
find_unused_bpf_program_index(descriptor_t *desc, unsigned long *index_p)
{
    int i;

    for (i = 0; i < 32; i++) {
        if (desc->valmap_bpf_program[i] == NULL) {
            *index_p = i;
            return 0;
        }
    }
    return -1;
}

static void
cleanup_valmap_bpf_program_index(descriptor_t *desc, int i, int do_cleanup_func)
{
    edtk_debug("%s: i = %d, do_cleanup_func = %d", __FUNCTION__, i, do_cleanup_func);
    pthread_mutex_lock(&desc->mutex);
    if (desc->valmap_bpf_program_flags[i] & VALMAP_INUSE) {
        edtk_debug("%s: i = %d, in use, setting VALMAP_DELAYED_CLEANUP flag", __FUNCTION__, i);
        desc->valmap_bpf_program_flags[i] |= VALMAP_DELAYED_CLEANUP;
        pthread_mutex_unlock(&desc->mutex);
        return;
    }
    desc->valmap_bpf_program_flags[i] = 0;
    pthread_mutex_unlock(&desc->mutex);
    if (do_cleanup_func) {
        edtk_debug("%s: calling func my_pcap_freecode", __FUNCTION__);
        my_pcap_freecode(desc->valmap_bpf_program[i]);
    }
    desc->valmap_bpf_program[i] = NULL;
}

static void
cleanup_valmap_bpf_program_all(descriptor_t *desc, int do_cleanup_func)
{
    int i;

    edtk_debug("%s: do_cleanup_func = %d", __FUNCTION__, do_cleanup_func);
    for (i = 0; i < 32; i++) {
        cleanup_valmap_bpf_program_index(desc, i, do_cleanup_func);
    }
}


